"""awaitables module tests"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
from asyncio import Task, Future
from typing import Awaitable

from pytest import raises

from dev4py.utils import awaitables
from dev4py.utils.types import Function, SyncOrAsync


class TestIsAwaitable:
    """is_awaitable function tests"""

    class TestNominalCase:
        def test_none__should__return_false(self) -> None:
            """When value is None should return False"""
            # GIVEN / WHEN
            result: bool = awaitables.is_awaitable(None)

            # THEN
            assert not result

        async def test_coroutine_value__should__return_true(self) -> None:
            """When value is a Coroutine should return True"""

            # GIVEN
            async def coroutine() -> None: ...

            value = coroutine()

            # WHEN
            result: bool = awaitables.is_awaitable(value)

            # THEN
            assert result
            await value  # Remove warning

        async def test_task_value__should__return_true(self) -> None:
            """When value is a Task should return True"""

            # GIVEN
            async def coroutine() -> None: ...

            value: Task = asyncio.create_task(coroutine())

            # WHEN
            result: bool = awaitables.is_awaitable(value)

            # THEN
            assert result
            await value  # Remove warning

        async def test_future_value__should__return_true(self) -> None:
            """When value is a Future should return True"""

            # GIVEN
            async def coroutine() -> None: ...

            value: Future = asyncio.ensure_future(coroutine())

            # WHEN
            result: bool = awaitables.is_awaitable(value)

            # THEN
            assert result
            await value  # Remove warning

        def test_not_awaitable_value__should__return_false(self) -> None:
            """When value is not an awaitable should return False"""
            # GIVEN / WHEN
            result: bool = awaitables.is_awaitable("A string")

            # THEN
            assert not result


class TestToSyncOrAsyncParamFunction:
    """to_sync_or_async_param_function function tests"""

    class TestNominalCase:

        async def test_sync_param_function__should__return_sync_or_async_param_async_function(self) -> None:
            """When given Function uses sync parameter, should return an async function with SyncOrAsync parameter"""
            # GIVEN
            param_value: str = "A string"
            sync_param_function: Function[str, str] = lambda s: s + '_suffix'

            # WHEN
            result_function: Function[SyncOrAsync[str], Awaitable[str]] = \
                awaitables.to_sync_or_async_param_function(sync_param_function)

            # THEN
            # -> Sync parameter call should work
            assert await result_function(param_value) == sync_param_function(param_value)

            # -> Async parameter call should work
            async def async_param_supplier() -> str: return param_value

            assert await result_function(async_param_supplier()) == sync_param_function(param_value)

        async def test_sync_param_async_function__should__return_sync_or_async_param_async_function(self) -> None:
            """When the given Function uses sync parameter and is async (i.e. returns an Awaitable[R]), should return an
            async function with SyncOrAsync parameter with Awaitable[R] result (not an Awaitable[Awaitable[R]])"""
            # GIVEN
            param_value: str = "A string"

            async def async_param_async_function(s: str) -> str: return s + '_suffix'

            # WHEN
            result_function: Function[SyncOrAsync[str], Awaitable[str]] = \
                awaitables.to_sync_or_async_param_function(async_param_async_function)

            # THEN
            # -> Sync parameter call should work
            assert await result_function(param_value) == await async_param_async_function(param_value)

            # -> Async parameter call should work
            async def async_param_supplier() -> str: return param_value

            assert await result_function(async_param_supplier()) == await async_param_async_function(param_value)

    class TestErrorCase:

        def test_none_function__should__raise_type_error(self) -> None:
            """When the function is None, should raise TypeError"""

            # GIVEN / WHEN / THEN
            with raises(TypeError):
                awaitables.to_sync_or_async_param_function(None)

        async def test_async_param_function__should__probably_raise_an_error(self) -> None:
            """When the given Function uses async parameter (i.e. Awaitable[T]), should probably raise an error because
            the wrapper will await the parameter before calling the given function"""

            # GIVEN
            async def async_param_supplier() -> str: return "A string"

            def async_param_function(awaitable: Awaitable[str]) -> str:
                # Do something...
                if not awaitables.is_awaitable(awaitable):
                    raise TypeError()
                return "A result string"

            # WHEN
            result_function: Function[SyncOrAsync[str], Awaitable[str]] = \
                awaitables.to_sync_or_async_param_function(async_param_function)

            # THEN
            async_value: Awaitable[str] = async_param_supplier()
            assert async_param_function(async_value) == "A result string"
            with raises(TypeError):
                await result_function(async_param_supplier())

            await async_value  # Remove warning

        async def test_async_param_async_function__should__probably_raise_an_error(self) -> None:
            """When the given Function uses async parameter (i.e. Awaitable[T]) and is async, should probably raise an
            error because the wrapper will await the parameter before calling the given function"""

            # GIVEN
            async def async_param_supplier() -> str: return "A string"

            async def async_param_async_function(awaitable: Awaitable[str]) -> str:
                return await awaitable

            # WHEN
            result_function: Function[SyncOrAsync[str], Awaitable[str]] = \
                awaitables.to_sync_or_async_param_function(async_param_async_function)

            # THEN
            assert await async_param_async_function(async_param_supplier()) == await async_param_supplier()
            with raises(TypeError):
                await result_function(async_param_supplier())


class TestToAwaitable:
    """to_awaitable function tests"""

    class TestNominalCase:
        async def test_none__should__return_awaitable_of_none(self) -> None:
            """When value is None should return Awaitable[None]"""
            # GIVEN / WHEN
            result: Awaitable[None] = awaitables.to_awaitable(None)

            # THEN
            assert await result is None

        async def test_t_coroutine_value__should__return_t_awaitable(self) -> None:
            """When value is a Coroutine of T should return Awaitable[T]"""

            # GIVEN
            expected_result: int = 1

            async def coroutine() -> int:
                return expected_result

            value = coroutine()

            # WHEN
            result: Awaitable[int] = awaitables.to_awaitable(value)

            # THEN
            assert await result == expected_result

        async def test_t_task_value__should__return_t_awaitable(self) -> None:
            """When value is a Task of T should return Awaitable[T]"""

            # GIVEN
            expected_result: int = 1

            async def coroutine() -> int:
                return expected_result

            value: Task = asyncio.create_task(coroutine())

            # WHEN
            result: Awaitable[int] = awaitables.to_awaitable(value)

            # THEN
            assert await result == expected_result

        async def test_t_future_value__should__return_t_awaitable(self) -> None:
            """When value is a Future of T should return Awaitable[T]"""

            # GIVEN
            expected_result: int = 1

            async def coroutine() -> int:
                return expected_result

            value: Future = asyncio.ensure_future(coroutine())

            # WHEN
            result: Awaitable[int] = awaitables.to_awaitable(value)

            # THEN
            assert await result == expected_result

        async def test_not_awaitable_t_value__should__return_t_awaitable(self) -> None:
            """When value of T is not an awaitable should return Awaitable[T]"""
            # GIVEN
            value: int = 1

            # WHEN
            result: Awaitable[int] = awaitables.to_awaitable(value)

            # THEN
            assert await result == value


class TestAsyncNone:
    """async_none function tests"""

    class TestNominalCase:
        async def test_should__return_awaitable_of_none(self) -> None:
            """Should return Awaitable[None]"""
            # GIVEN / WHEN
            result: Awaitable[None] = awaitables.async_none()

            # THEN
            assert await result is None



"""The `retry` module provides tools to create retryable callable using exponential backoff"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from asyncio import sleep as async_sleep
from dataclasses import dataclass
from functools import partial
from time import sleep
from typing import Callable, Awaitable

from dev4py.utils.objects import require_non_none
from dev4py.utils.types import T, P, Function


##############################
#     PUBLIC MODULE CLASS    #
##############################
@dataclass(frozen=True)
class RetryConfiguration:
    """
    RetryConfiguration class that represents a retry configuration

    A retry configuration describes:
        * The maximum number of tries (first try included. I.e.: 1 means no retry on failure)
        * The exponential backoff elements (exponent & delay)
            Note: exponential backoff (/ waiting_interval) = delay * (exponent^retry_number)

    Args:
        exponent (int): the exponential backoff exponent / an arbitrary multiplier to determine delay between each try
            (default = 2)
        delay (float): the exponential backoff delay in second / the initial wait for the first retry (default = 0.1)
        max_tries (int): max try number (first try included) (default = 3, i.e.: first try and 2 retry)

    Raises:
        TypeError: if exponent or delay or max_tries is None
        ValueError: if exponent or delay is less than 0 or max_tries is less than 1
    """
    exponent: int = 2
    delay: float = 0.1  # seconds
    max_tries: int = 3

    def get_waiting_interval(self, retry_number: int) -> float:
        """
        Returns the waiting interval in second in case of a retry occurs
        Args:
            retry_number: the current retry number

        Returns:
            float: the waiting interval in second for the given `retry_numbers`

        Raises:
            TypeError: if retry_number is None
            ValueError: if retry_number is lesser than or equals to 0 or retry_number greater than max_tries
        """
        if require_non_none(retry_number) <= 0:
            raise ValueError('retry_number must be greater than 0')

        if retry_number > self.max_tries:
            raise ValueError('retry_number greater than configured max_tries')

        return self.delay * (self.exponent ** retry_number)

    def __post_init__(self):
        if require_non_none(self.exponent, "exponent must be non None") < 0:
            raise ValueError("exponent must be greater than or equals to 0")
        if require_non_none(self.delay, "delay must be non None") < 0:
            raise ValueError("delay must be greater than or equals to 0")
        if require_non_none(self.max_tries, "max_tries must be non None") < 1:
            raise ValueError("max_tries must be greater than or equals to 1")


##############################
#  PRIVATE MODULE FUNCTIONS  #
##############################
def _default_retry_on_failure(exception: BaseException) -> T:
    """
    Default value for on_failure parameter
    Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)

    Args:
        exception: The on failure exception

    Raises:
        Exception: the given exception
    """
    raise exception


def _with_retry(
        sync_callable: Callable[P, T],
        retry_config: RetryConfiguration,
        on_failure: Function[BaseException, T],
        retry_number: int,
        *args: P.args,
        **kwargs: P.kwargs
) -> T:
    """
    Used to define a retryable function from `sync_callable` parameter by using the RetryConfig, on_failure and
    retry_number parameters

    Note: inner functions are not used in order to be compatible with multiprocessing
    """
    if retry_number > 0:
        sleep(retry_config.get_waiting_interval(retry_number))

    try:
        return sync_callable(*args, **kwargs)
    except BaseException as e:  # pylint: disable=W0703
        retry_number += 1  # pragma: no mutate

        if retry_number >= retry_config.max_tries:
            return on_failure(e)

        return _with_retry(sync_callable, retry_config, on_failure, retry_number, *args, **kwargs)


async def _with_async_retry(
        async_callable: Callable[P, Awaitable[T]],
        retry_config: RetryConfiguration,
        on_failure: Function[BaseException, T],
        retry_number: int,
        *args: P.args,
        **kwargs: P.kwargs
) -> T:
    """
    Used to define a retryable function from `async_callable` parameter by using the RetryConfig, on_failure and
    retry_number parameters

    Note: inner functions are not used in order to be compatible with multiprocessing
    """
    if retry_number > 0:
        await async_sleep(retry_config.get_waiting_interval(retry_number))

    try:
        return await async_callable(*args, **kwargs)
    except BaseException as e:  # pylint: disable=W0703
        retry_number += 1  # pragma: no mutate

        if retry_number >= retry_config.max_tries:
            return on_failure(e)

        return await _with_async_retry(async_callable, retry_config, on_failure, retry_number, *args, **kwargs)


##############################
#      PUBLIC FUNCTIONS      #
##############################
def to_retryable(
        sync_callable: Callable[P, T],
        retry_config: RetryConfiguration = RetryConfiguration(),
        on_failure: Function[BaseException, T] = _default_retry_on_failure
) -> Callable[P, T]:
    """
    Transforms a callable to a retryable one by using the given RetryConfiguration

    If an error occurs when executing the returned callable the call is retried by using the given RetryConfiguration.
    An exponential backoff is used in order to determinate the waiting time between each call. If max_tries is reached
    then the on_failure function is called in order to return a default value or raise an exception. By default, it
    raises the last raised exception

    Args:
        sync_callable: The callable to transform
        retry_config: The given RetryConfiguration
        on_failure: The action to perform if max_tries is reached (Note: can return a default value)

    Returns:
        Callable[P, T]: The retryable callable

    Raises:
        TypeError: if sync_callable or retry_config or on_failure is None
    """
    require_non_none(sync_callable)
    require_non_none(retry_config)
    require_non_none(on_failure)
    return partial(_with_retry, *[sync_callable, retry_config, on_failure, 0])


def to_async_retryable(
        async_callable: Callable[P, Awaitable[T]],
        retry_config: RetryConfiguration = RetryConfiguration(),
        on_failure: Function[BaseException, T] = _default_retry_on_failure
) -> Callable[P, Awaitable[T]]:
    """
    Transforms an async callable to an async retryable one by using the given RetryConfiguration

    If an error occurs when executing the returned callable the call is retried by using the given RetryConfiguration.
    An exponential backoff is used in order to determinate the waiting time between each call. If max_tries is reached
    then the on_failure function is called in order to return a default value or raise an exception. By default, it
    raises the last raised exception

    Args:
        async_callable: The async callable to transform
        retry_config: The given RetryConfiguration
        on_failure: The action to perform if max_tries is reached (Note: can return a default value)

    Returns:
        Callable[P, Awaitable[T]]: The async retryable callable

    Raises:
        TypeError: if async_callable or retry_config or on_failure is None
    """
    require_non_none(async_callable)
    require_non_none(retry_config)
    require_non_none(on_failure)
    return partial(_with_async_retry, *[async_callable, retry_config, on_failure, 0])



"""Stream module inspired by `java.util.stream`"""  # pylint: disable=C0302

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from concurrent.futures import Executor, wait, FIRST_COMPLETED, Future, as_completed
from dataclasses import dataclass
from functools import partial, cmp_to_key
from typing import Generic, Final, Optional, cast, Any, Collection, Iterable, Iterator

from dev4py.utils import collectors
from dev4py.utils.JOptional import JOptional
from dev4py.utils.collectors import Collector
from dev4py.utils.iterables import get_chunks
from dev4py.utils.objects import require_non_none, require_non_none_else_get, to_self
from dev4py.utils.pipeline import StepPipeline, StepResult
from dev4py.utils.types import T, Function, R, V, Predicate, Supplier, BiConsumer, K, Consumer, BiFunction


##############################
#  PRIVATE MODULE FUNCTIONS  #
##############################
def _default_handler(v: V) -> StepResult[R]:
    """
    private function to describe default pipeline handler
    Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
    """
    # lambda v: StepResult(cast(R, v))
    return StepResult(cast(R, v))


def _root_pipeline() -> StepPipeline[V, R]:
    """
    private function to describe default pipeline
    Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
    """
    return StepPipeline.of(_default_handler)


def _none_collector() -> Collector[T, R]:
    """
    private function to cast none_collector
    Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
    """
    return cast(Collector[T, R], collectors.to_none())


def _map_lambda(v: T, mapper: Function[T, R]) -> StepResult[R]:
    """
    private function to describe map method handler
    Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
    """
    # lambda v: StepResult(value=mapper(v))
    return StepResult(value=mapper(v))


def _filter_lambda(v: T, predicate: Predicate[T]) -> StepResult[T]:
    """
    private function to describe filter method handler
    Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
    """
    # lambda v: StepResult(value=v, go_next=predicate(v))
    return StepResult(value=v, go_next=predicate(v))


def _and_lambda(b1: bool, b2: bool) -> bool:
    """
    private function to describe a AND BiFunction
    Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
    """
    # lambda b1, b2: b1 and b2
    return b1 and b2


def _stream_to_list_lambda(stream: Stream[T]) -> list[T]:
    """
    private function to describe a stream to list Function
    Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
    Note: use lists because generators cannot be used in parallel context (generators are not serializable)
    """
    # lambda stream: stream.to_list()
    return stream.to_list()


def _peek_mapper(value: T, consumer: Consumer[T]) -> T:
    """
    private function to describe peek method mapper
    Note: inner function are not used in order to be compatible with multiprocessing (inner function are not
    serializable)
    """
    consumer(value)
    return value


def _natural_comparator(o1: Any, o2: Any) -> int:
    """
    private function to describe a natural order comparator
    Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
    """
    return 0 if o1 == o2 else 1 if o1 > o2 else -1  # pragma: no mutate


def _to_max_comparator(o1: T, o2: T, comparator: BiFunction[T, T, int]) -> int:
    """
    private function to describe a reverse order comparator
    Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
    """
    return -comparator(o1, o2)


def _sync_execution(
        values: Iterable[V],
        pipeline: StepPipeline[V, T],
        stop_on_first_completed: bool,
        collector: Collector[T, R]
) -> R:
    """
    private function to describe a sync stream execution
    Note: Outside the Stream class in order to be compatible with multiprocessing
    """
    return _process_values(values, pipeline, stop_on_first_completed, collector)[0]


def _parallel_execution(  # pylint: disable=R0913
        values: Iterable[V],
        pipeline: StepPipeline[V, T],
        parallel_config: ParallelConfiguration,
        ordered_execution: bool,
        stop_on_first_completed: bool,
        collector: Collector[T, R]
) -> R:
    """
    private function to describe a parallel stream execution
    Note: Outside the Stream class in order to be compatible with multiprocessing
    """
    chunksize: int = parallel_config.chunksize
    executor: Executor = parallel_config.executor
    process_values: Final[partial[tuple[R, bool]]] = \
        partial(
            _process_values, pipeline=pipeline, stop_on_first_completed=stop_on_first_completed, collector=collector
        )
    futures: Final[list[Future[tuple[R, bool]]]] = \
        [executor.submit(process_values, chunk) for chunk in get_chunks(values, chunksize)]

    return (_ordered_parallel_execution if ordered_execution else _unordered_parallel_execution)(
        futures=futures,
        collector=collector
    )


def _unordered_parallel_execution(futures: list[Future[tuple[R, bool]]], collector: Collector[T, R]) -> R:
    """
    private function to describe an unordered parallel stream execution
    Note: Outside the Stream class in order to be compatible with multiprocessing
    """
    done: Collection[Future[tuple[R, bool]]]
    not_done: Collection[Future[tuple[R, bool]]] = futures

    try:
        results: R = collector.supplier()
        continue_processing: bool = True

        # 'do {...} while' doesn't exist in python :'(
        while not_done and (continue_processing is True):  # pragma: no mutate
            done, not_done = wait(not_done, return_when=FIRST_COMPLETED)
            for future in done:
                chunk_result: tuple[R, bool] = future.result()
                results = collector.combiner(results, chunk_result[0])
                continue_processing = chunk_result[1]
                if continue_processing is False:
                    break  # pragma: no mutate

        return results
    finally:
        for future in not_done:
            future.cancel()


def _ordered_parallel_execution(futures: list[Future[tuple[R, bool]]], collector: Collector[T, R]) -> R:
    """
    private function to describe an ordered parallel stream execution
    Note: Outside the Stream class in order to be compatible with multiprocessing
    """
    results: R = collector.supplier()
    processed_value: int = 0  # pragma: no mutate
    try:
        for future in futures:
            processed_value += 1  # pragma: no mutate
            chunk_result: tuple[R, bool] = future.result()
            results = collector.combiner(results, chunk_result[0])
            if chunk_result[1] is False:
                break  # pragma: no mutate
    finally:
        for i in range(processed_value, len(futures)):
            futures[i].cancel()

    return results


def _process_values(
        values: Iterable[V],
        pipeline: StepPipeline[V, T],
        stop_on_first_completed: bool,
        collector: Collector[T, R]
) -> tuple[R, bool]:
    """
    private function to describe process stream value set execution
    Note: Outside the Stream class in order to be compatible with multiprocessing
    """
    continue_processing: bool = True
    # Note: Always ordered
    results: R = collector.supplier()
    for result in _sync_generator(values, pipeline):
        results = collector.accumulator(results, result)
        if stop_on_first_completed:
            continue_processing = False  # pragma: no mutate
            break  # pragma: no mutate

    return results, continue_processing


def _sync_generator(values: Iterable[V], pipeline: StepPipeline[V, T]) -> Iterable[T]:
    """
    private function to describe a sync Stream generator
    Note: Outside the Stream class in order to be compatible with multiprocessing
    """
    for value in values:
        result: StepResult[Any] = pipeline.execute(value)
        # Note: go_next TRUE means the value has gone through the whole pipeline and has not been filtered
        if result.go_next:
            yield result.value


def _parallel_generator(
        values: Iterable[V],
        pipeline: StepPipeline[V, T],
        parallel_config: ParallelConfiguration,
        ordered_execution: bool
) -> Iterable[T]:
    """
    private function to describe a parallel Stream generator (ordered or not)
    Note: Outside the Stream class in order to be compatible with multiprocessing
    """
    chunksize: int = parallel_config.chunksize
    executor: Executor = parallel_config.executor
    process_values: Final[partial[tuple[list[T], bool]]] = \
        partial(_process_values, pipeline=pipeline, stop_on_first_completed=False, collector=collectors.to_list())
    futures: Final[list[Future[tuple[list[T], bool]]]] = \
        [executor.submit(process_values, chunk) for chunk in get_chunks(values, chunksize)]
    try:
        futures_iterable: Final[Iterable[Future[tuple[list[T], bool]]]] = \
            (futures if ordered_execution else as_completed(futures))
        for future in futures_iterable:
            for value in future.result()[0]:
                yield value

    except BaseException as e:
        for future in futures:
            future.cancel()
        raise e


##############################
#       PUBLIC CLASSES       #
##############################
@dataclass(frozen=True)
class ParallelConfiguration:
    """
    ParallelConfiguration class that describes Stream parallel configuration when using parallel execution

    Args:
        executor (Executor): The executor to use for parallel (/concurrent) execution
        chunksize (int): If greater than one, the stream values will be chopped into chunks of size chunksize and
            submitted to the process pool. If set to one, the items in the stream will be sent one at a time.

    Raises:
        TypeError: if executor is None
        ValueError: if chunksize is less than 1
    """
    executor: Executor
    chunksize: int = 1

    def __post_init__(self):
        require_non_none(self.executor, "executor must be non None")
        if require_non_none(self.chunksize, "chunksize must be non None") < 1:
            raise ValueError("chunksize must be greater than or equals to 1")


class Stream(Generic[T]):  # pylint: disable=R0904
    __CREATE_KEY: Final[object] = object()

    @classmethod
    def empty(cls) -> Stream[T]:
        """
        Returns an empty sequential Stream

        Returns:
            Stream[T]: An empty sequential Stream
        """
        return cls.of()

    @classmethod
    def of(cls, *values: T) -> Stream[T]:
        """
        Returns a sequential ordered stream whose elements are the specified values

        Args:
            *values: ordered stream elements

        Returns:
            Stream[T]: A sequential ordered stream whose elements are the specified values
        """
        return cls.of_iterable(values)

    @classmethod
    def of_iterable(cls, iterable: Iterable[T]) -> Stream[T]:
        """
        Returns a sequential ordered stream whose elements are values from the given Iterable

        Args:
            iterable: iterable of stream source values

        Returns:
            Stream[T]: A sequential ordered stream whose elements are values from the given Iterable

        Raises:
            TypeError: if iterable is None
        """
        require_non_none(iterable)
        return cls._of(lambda p, o: iterable)

    @classmethod
    def _of(
            cls,
            values_function: BiFunction[Optional[ParallelConfiguration], bool, Iterable[V]],
            pipeline: Optional[StepPipeline[V, R]] = None,
            parallel_config: Optional[JOptional[ParallelConfiguration]] = None,
            ordered_execution: bool = False
    ) -> Stream[R]:
        """
        Private class method in order to simplify the call to Stream private constructor

        Args:
            values_function: The BiFunction which provides the Stream values
            pipeline: The StepPipeline to be executed if presents by the Stream values on terminal operation
            parallel_config: The Stream parallel configuration
            ordered_execution: TRUE if the value must be returned to the encounter order

        Returns:
            Stream[T]: A Stream corresponding to the given parameters
        """
        return Stream(
            values_function=values_function,
            pipeline=require_non_none_else_get(pipeline, _root_pipeline),
            parallel_config=require_non_none_else_get(parallel_config, JOptional.empty),
            ordered_execution=ordered_execution,
            create_key=cls.__CREATE_KEY
        )

    def __init__(  # pylint: disable=R0913
            self,
            values_function: BiFunction[Optional[ParallelConfiguration], bool, Iterable[V]],
            pipeline: StepPipeline[V, T],
            parallel_config: JOptional[ParallelConfiguration],
            ordered_execution: bool,
            create_key: object,
    ):
        """Stream private constructor: Constructs a Stream[T] inspired by java `java.util.stream.Stream<T>`"""
        assert create_key == self.__CREATE_KEY, "Stream private constructor! Please use Stream.of"
        self._values_function: BiFunction[Optional[ParallelConfiguration], bool, Iterable[V]] = \
            require_non_none(values_function)
        self._pipeline: StepPipeline[V, T] = require_non_none(pipeline)
        self._parallel_config: JOptional[ParallelConfiguration] = require_non_none(parallel_config)
        self._ordered_execution: bool = require_non_none(ordered_execution)

    def map(self, mapper: Function[T, R]) -> Stream[R]:
        """
        Returns a stream consisting of the results of applying the given function to the elements of this stream

        Args:
            mapper (Function[T, R]): The function to apply to each element

        Returns:
            Stream[R]: A stream consisting of the results of applying the given function to the elements of this stream

        Raises:
            TypeError: if mapper is None
        """
        require_non_none(mapper)
        return Stream._of(
            values_function=self._values_function,
            # pylint: disable=E1101
            pipeline=self._pipeline.add_handler(cast(Function[T, StepResult[R]], partial(_map_lambda, mapper=mapper))),
            parallel_config=self._parallel_config,
            ordered_execution=self._ordered_execution
        )

    def parallel(self, parallel_config: Optional[ParallelConfiguration]) -> Stream[T]:
        """
        Configures the Stream to use the given parallel configuration and return the current stream

        Note: None parallel configuration is like calling `sequential()` method

        Args:
            parallel_config: The parallel configuration to use

        Returns:
            Stream[T]: The current Stream after set parallel configuration
        """
        self._parallel_config = JOptional.of_noneable(parallel_config)
        return self

    def is_parallel(self) -> bool:
        """
        Returns whether this stream, if a terminal operation were to be executed, would execute in parallel

        Returns:
            bool: True if the Stream uses parallel configuration, otherwise False
        """
        return self._parallel_config.is_present()

    def sequential(self) -> Stream[T]:
        """
        Configures the Stream to use sequential configuration and return the current stream
        (i.e.: Remove the parallel configuration if exists)

        Returns:
            Stream[T]: The current Stream after set sequential configuration
        """
        if self.is_parallel():
            self._parallel_config = JOptional.empty()
        return self

    def filter(self, predicate: Predicate[T]) -> Stream[T]:
        """
        Returns a stream consisting of the elements of this stream that match the given predicate

        Args:
            predicate: The predicate to apply to each element to determine if it should be included

        Returns:
            Stream[T]: A stream consisting of the elements of this stream that match the given predicate

        Raises:
            TypeError: if predicate is None
        """
        return Stream._of(
            values_function=self._values_function,
            # pylint: disable=E1101
            pipeline=self._pipeline.add_handler(
                cast(Function[T, StepResult[T]], partial(_filter_lambda, predicate=require_non_none(predicate)))
            ),
            parallel_config=self._parallel_config,
            ordered_execution=self._ordered_execution
        )

    def collect(self, collector: Collector[T, R]) -> R:
        """
        Performs a reduction operation on the elements of this stream using a Collector

        Args:
            collector: The Collector describing the reduction

        Returns:
            R: The result of the reduction

        Raises:
            TypeError: if collector is None
        """
        return self._execute(collector=require_non_none(collector))

    def collect_from(self, supplier: Supplier[R], accumulator: BiConsumer[R, T], combiner: BiConsumer[R, R]) -> R:
        """
        Performs a reduction operation on the elements of this stream

        Args:
            supplier: A function that creates a new result container. For a parallel execution, this function may be
                called multiple times and must return a fresh value each time
            accumulator: A function that must fold an element into a result container
            combiner: A unction that accepts two partial result containers and merges them, which must be compatible
                with the accumulator function. The combiner function must fold the elements from the second result
                container into the first result container

        Returns:
            R: The result of the reduction

        Raises:
            TypeError: if at least one parameter is None
        """
        return self.collect(collector=collectors.of_biconsumers(supplier, accumulator, combiner))

    def to_list(self) -> list[T]:
        """
        Accumulates the elements of this stream into a list

        Returns:
            list[T]: A list containing the stream elements
        """
        return self.collect(collector=collectors.to_list())

    def to_tuple(self) -> tuple[T, ...]:
        """
        Accumulates the elements of this stream into a tuple

        Returns:
            tuple[T, ...]: a tuple containing the stream elements
        """
        return self.collect(collector=collectors.to_tuple())

    def to_dict(self, key_mapper: Function[T, K], value_mapper: Function[T, V]) -> dict[K, V]:
        """
        Accumulates the elements of this stream into a tuple dict whose keys and values are the result of applying the
        provided mapping functions to the input elements

        Args:
            key_mapper: a mapping function to produce keys
            value_mapper: a mapping function to produce values

        Returns:
            dict[K, V]: a dict containing the stream elements whose keys and values are the result of applying mapping
                functions to the input elements
        """
        return self.collect(collector=collectors.to_dict(key_mapper, value_mapper))

    def all_match(self, predicate: Predicate[T]) -> bool:
        """
        Returns whether all elements of this stream match the provided predicate

        Args:
            predicate: The predicate to apply to elements of this stream

        Returns:
            bool: True if either all elements of the stream match the provided predicate or the stream is empty,
                otherwise False

        Raises:
            TypeError: if predicate is None
        """
        return self.map(predicate).reduce(True, _and_lambda)

    def any_match(self, predicate: Predicate[T]) -> bool:
        """
        Returns whether any elements of this stream match the provided predicate

        Args:
            predicate: The predicate to apply to elements of this stream

        Returns:
            bool: True if any elements of the stream match the provided predicate, otherwise False

        Raises:
            TypeError: if predicate is None
        """
        return self.filter(predicate).find_any().is_present()

    def count(self) -> int:
        """
        Returns the count of elements in this stream

        Returns:
            int: The count of elements in this stream
        """
        return self.collect(collector=collectors.to_counter())

    def for_each(self, consumer: Consumer[T]) -> None:
        """
        Performs an action (/consumer) for each element of this stream

        Args:
            consumer: The Consumer to perform on the elements

        Returns:
            None: nothing

        Raises:
            TypeError: if consumer is None
        """
        self.peek(require_non_none(consumer))
        self._execute()

    def flat_map(self, mapper: Function[T, Stream[R]]) -> Stream[R]:
        """
        Returns a stream consisting of the results of replacing each element of this stream with the contents of a
        mapped stream produced by applying the provided mapping function to each element

        Args:
            mapper: A function to apply to each element which produces a stream of new values

        Returns:
             Stream[R]: The new Stream

        Raises:
            TypeError: if mapper is None
        """
        return self._from_self_config(
            values_function=cast(
                BiFunction[Optional[ParallelConfiguration], bool, Iterable[R]],
                partial(self._flat_map_values_function, mapper=require_non_none(mapper))
            )
        )

    def ordered_execution(self, ordered: bool = True) -> Stream[T]:
        """
        Configures the current Stream to be ordered or not and returns it

        Note: This configuration has no impact in sequential stream

        Args:
            ordered: True if the Stream must be ordered, False for unordered. (Default: True)

        Returns:
            Stream[T]: The current Stream with ordered configuration

        Raises:
            TypeError: if ordered is None
        """
        return Stream._of(
            values_function=self._values_function,
            pipeline=self._pipeline,
            parallel_config=self._parallel_config,
            ordered_execution=ordered
        ) if self._ordered_execution != require_non_none(ordered) else self

    def unordered(self) -> Stream[T]:
        """
        Configures the current Stream to be unordered and returns it

        Note: Is equivalent to call `ordered_execution(False)`
        Note 2: This configuration has no impact in sequential stream

        Returns:
            Stream[T]: The current Stream with unordered configuration
        """
        return self.ordered_execution(False)

    def find_first(self) -> JOptional[T]:
        """
        Returns a JOptional describing the first element of this stream, or an empty JOptional if the stream is empty

        Returns:
            JOptional[T]: A JOptional describing the first element of this stream, or an empty JOptional if the stream
                is empty
        """
        return self.ordered_execution(True).find_any()

    def find_any(self) -> JOptional[T]:
        """
        Returns a JOptional describing some element of the stream, or an empty JOptional if the stream is empty

        Note: if the stream is ordered, find_any and find_first are equivalents

        Returns:
            JOptional[T]: A JOptional describing some element of the stream, or an empty JOptional if the stream is
                empty
        """
        results: list[T] = self._execute(
            stop_on_first_completed=True  # pragma: no mutate
            , collector=collectors.to_list()
        )
        return JOptional.of_noneable(results[0] if results else None)

    def distinct(self) -> Stream[T]:
        """
        Returns a stream consisting of the distinct elements of this stream

        Returns:
            Stream[T]: A stream consisting of the distinct elements of this stream
        """
        return self._from_self_config(values_function=self._distinct_values_function)

    def drop_while(self, predicate: Predicate[T]) -> Stream[T]:
        """
        Returns, if this stream is ordered, a stream consisting of the remaining elements of this stream after dropping
        the longest prefix of elements that match the given predicate. Otherwise, if this stream is unordered, returns a
        stream consisting of the remaining elements of this stream after dropping a subset of elements that match the
        given predicate.

        Args:
            predicate: The predicate to apply to elements to determine the longest prefix of elements

        Returns:
            Stream[T]: The new Stream

        Raises:
            TypeError: if predicate is None
        """
        return self._from_self_config(
            values_function=partial(self._drop_while_values_function, predicate=require_non_none(predicate))
        )

    def limit(self, limit: int) -> Stream[T]:
        """
        Returns a stream consisting of the elements of this stream, truncated to be no longer than maxSize in length

        Args:
            limit: The number of elements the stream should be limited to

        Returns:
            Stream[T]: The truncated Stream

        Raises:
            TypeError: if limit is None
        """
        return self._from_self_config(
            values_function=partial(self._limit_values_function, limit=require_non_none(limit)))

    def max(self, comparator: BiFunction[T, T, int] = _natural_comparator) -> JOptional[T]:
        """
        Returns the maximum element of this stream according to the provided comparator BiFunction

        Args:
            comparator: The comparator BiFunction to compare elements of this stream (Note: Returns a negative int,
                zero, or a positive integer as the first argument is less than, equal to, or greater than the second)
                (default: natural comparator)

        Returns:
            JOptional[T]: A JOptional describing the maximum element of this stream, or an empty JOptional if the stream
                is empty

        Raises:
            TypeError: if comparator is None
        """
        return self.min(partial(_to_max_comparator, comparator=require_non_none(comparator)))

    def min(self, comparator: BiFunction[T, T, int] = _natural_comparator) -> JOptional[T]:
        """
        Returns the minimum element of this stream according to the provided comparator BiFunction

        Args:
            comparator: The comparator BiFunction to compare elements of this stream (Note: Returns a negative int,
                zero, or a positive integer as the first argument is less than, equal to, or greater than the second)
                (default: natural comparator)

        Returns:
            JOptional[T]: A JOptional describing the minimum element of this stream, or an empty JOptional if the stream
                is empty

        Raises:
            TypeError: if comparator is None
        """
        return self.sorted(require_non_none(comparator)).find_first()

    def peek(self, consumer: Consumer[T]) -> Stream[T]:
        """
        Returns a stream consisting of the elements of this stream, additionally performing the provided action
        (/consumer) on each element as elements are consumed from the resulting stream

        Args:
            consumer: The Consumer to perform on the elements as they are consumed from the stream

        Returns:
            Stream[T]: The new Stream

        Raises:
            TypeError: if consumer is None
        """
        return self.map(cast(Function[T, T], partial(_peek_mapper, consumer=require_non_none(consumer))))

    def reduce(self, identity: T, accumulator: BiFunction[T, T, T]) -> T:
        """
        Performs a reduction on the elements of this stream, using the provided identity value and an associative
        accumulation function, and returns the reduced value

        Args:
            identity: The identity value for the accumulating function
            accumulator: The BiFunction for combining two values

        Returns:
            T: The result of the reduction

        Raises:
            TypeError: accumulator is None
        """
        return self.collect(collectors.of(supplier=partial(to_self, obj=identity),
                                          accumulator=accumulator,
                                          combiner=accumulator
                                          )
                            )

    def skip(self, n: int) -> Stream[T]:
        """
        Returns a stream consisting of the remaining elements of this stream after discarding the first n elements of
        the stream

        Note: If this stream contains fewer than n elements then an empty stream will be returned

        Args:
            n: The number of leading elements to skip

        Returns:
            Stream[T]: The new stream

        Raises:
            TypeError: if n is None

        """
        return self._from_self_config(values_function=partial(self._skip_values_function, n=require_non_none(n)))

    def take_while(self, predicate: Predicate[T]) -> Stream[T]:
        """
        Returns, if this stream is ordered, a stream consisting of the longest prefix of elements taken from this stream
        that match the given predicate. Otherwise, if this stream is unordered, returns a stream consisting of a subset
        of elements taken from this stream that match the given predicate

        Args:
            predicate: The predicate to apply to elements to determine the longest prefix of element

        Returns:
            Stream[T]: The new Stream

        Raises:
            TypeError: if predicate is None
        """
        return self._from_self_config(
            values_function=partial(self._take_while_values_function, predicate=require_non_none(predicate))
        )

    def sorted(self, comparator: BiFunction[T, T, int] = _natural_comparator) -> Stream[T]:
        """
        Returns a stream consisting of the elements of this stream, sorted according to the provided comparator

        Args:
            comparator: The comparator BiFunction to compare elements of this stream (Note: Returns a negative int,
                zero, or a positive integer as the first argument is less than, equal to, or greater than the second)
                (default: natural comparator)

        Returns:
            Stream[T]: The new Stream

        Raises:
            TypeError: if comparator is None
        """
        self._ordered_execution = True
        return self._from_self_config(
            values_function=partial(self._sorted_values_function, comparator=require_non_none(comparator))
        )

    def to_generator(self) -> Iterator[T]:
        """
        Returns a generator based on Stream inputs and mapping

        Note: If the Stream is sequential each value is processed when __next__() is called. Otherwise, if the Stream
        uses a parallel configuration, all values are processed at the first __next__() call but each value is returned
        one by one depending on the 'ordered/unordered' configuration

        Note 2: When stream uses parallel with unordered configuration each value is returned when it is available. With
        an ordered configuration the generator wait for each necessary value in order to respect the stream value order

        Returns:
            Iterator[T]: The Stream processed values generator
        """
        # pylint: disable=E1102
        values: Final[Iterable[Any]] = \
            self._values_function(self._parallel_config.or_else(None), self._ordered_execution)
        if not values:
            return

        if self.is_parallel():
            yield from _parallel_generator(
                values=values,
                pipeline=self._pipeline,
                parallel_config=self._parallel_config.get(),
                ordered_execution=self._ordered_execution
            )
        else:
            yield from _sync_generator(values=values, pipeline=self._pipeline)

    def _execute(
            self,
            stop_on_first_completed: bool = False,
            collector: Collector[T, R] = _none_collector()
    ) -> R:
        """
        private method to execute the current Stream
        """
        require_non_none(stop_on_first_completed)
        require_non_none(collector)
        # pylint: disable=E1102
        values: Final[Optional[Iterable[Any]]] = \
            self._values_function(self._parallel_config.or_else(None), self._ordered_execution)

        if not values:
            return collector.supplier()

        if self.is_parallel():
            return _parallel_execution(
                values=values,
                pipeline=self._pipeline,
                parallel_config=self._parallel_config.get(),
                ordered_execution=self._ordered_execution,
                stop_on_first_completed=stop_on_first_completed,
                collector=collector
            )

        return _sync_execution(
            values=values, pipeline=self._pipeline, stop_on_first_completed=stop_on_first_completed, collector=collector
        )

    def _from_self_config(
            self, values_function: BiFunction[Optional[ParallelConfiguration], bool, Iterable[V]]
    ) -> Stream[R]:
        """
        private method that returns a new Stream by copying the current configuration
        Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
        """
        return Stream._of(
            values_function=values_function,
            parallel_config=self._parallel_config,
            ordered_execution=self._ordered_execution
        )

    def _init_to_values_function(
            self, parallel_config: Optional[ParallelConfiguration], ordered_execution: bool
    ) -> Stream[T]:
        """
        private method that initializes the current stream with the given configuration
        Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
        """
        return self.ordered_execution(ordered=ordered_execution).parallel(parallel_config)

    def _sorted_values_function(
            self, parallel_config: ParallelConfiguration, ordered_execution: bool, comparator: BiFunction[T, T, int]
    ) -> Iterable[T]:
        """
        private method used by `sorted` public method in replacement of lambda
        Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
        """
        values: list[T] = self._init_to_values_function(parallel_config, ordered_execution).to_list()
        values.sort(key=cmp_to_key(comparator))
        return values

    def _flat_map_values_function(
            self, parallel_config: Optional[ParallelConfiguration],
            ordered_execution: bool,
            mapper: Function[T, Stream[R]]
    ) -> Iterable[R]:
        """
        private method used by `flat_map` public method in replacement of lambda
        Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
        """
        list_generator: Final[Iterable[list[R]]] = self \
            ._init_to_values_function(parallel_config, ordered_execution) \
            .map(mapper) \
            .map(_stream_to_list_lambda) \
            .to_generator()
        for values in list_generator:
            for value in values:
                yield value

    def _distinct_values_function(
            self, parallel_config: Optional[ParallelConfiguration], ordered_execution: bool
    ) -> Iterable[T]:
        """
        private method used by `distinct` public method in replacement of lambda
        Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
        """
        existing_values: set[T] = set()
        for value in self._init_to_values_function(parallel_config, ordered_execution).to_generator():
            if not value in existing_values:
                existing_values.add(value)
                yield value

    def _drop_while_values_function(
            self, parallel_config: ParallelConfiguration, ordered_execution: bool, predicate: Predicate[T]
    ) -> Iterable[T]:
        """
        private method used by `drop_while` public method in replacement of lambda
        Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
        """
        drop: bool = True
        for value in self._init_to_values_function(parallel_config, ordered_execution).to_generator():
            drop = drop and predicate(value)
            if drop is False:
                yield value

    def _limit_values_function(
            self, parallel_config: ParallelConfiguration, ordered_execution: bool, limit: int
    ) -> Iterable[T]:
        """
        private method used by `limit` public method in replacement of lambda
        Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
        """
        if limit <= 0:
            return

        counter: int = 0
        for value in self._init_to_values_function(parallel_config, ordered_execution).to_generator():
            yield value
            counter += 1
            if counter >= limit:
                break

    def _skip_values_function(
            self, parallel_config: ParallelConfiguration, ordered_execution: bool, n: int
    ) -> Iterable[T]:
        """
        private method used by `skip` public method in replacement of lambda
        Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
        """
        counter: int = 0
        for value in self._init_to_values_function(parallel_config, ordered_execution).to_generator():
            if counter >= n:
                yield value
            counter += 1

    def _take_while_values_function(
            self, parallel_config: ParallelConfiguration, ordered_execution: bool, predicate: Predicate[T]
    ) -> Iterable[T]:
        """
        private method used by `take_while` public method in replacement of lambda
        Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
        """
        for value in self._init_to_values_function(parallel_config, ordered_execution).to_generator():
            if not predicate(value):
                break  # pragma: no mutate
            yield value



# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

from dev4py.utils.objects import require_non_none
from dev4py.utils.types import T


def empty_tuple() -> tuple[Any, ...]:
    """
    Returns an empty tuple

    Returns:
        tuple[Any, ...]: An empty tuple
    """
    return tuple()


def append(tpl: tuple[T, ...], value: T) -> tuple[T, ...]:
    """
    Adds the given value to the given tuple and returns the new tuple with added value

    Args:
        tpl: the tuple
        value: the value to add

    Returns:
        tuple[T, ...]: A new tuple with all elements from tpl and the added value

    Raises:
        TypeError: if the tuple is None
    """
    return extend(tpl, (value,))


def extend(tpl_1: tuple[T, ...], tpl_2: tuple[T, ...]) -> tuple[T, ...]:
    """
    Adds all values from the second tuple to the first one and returns the new created tuple

    Args:
        tpl_1: The tuple where to add elements
        tpl_2: The tuple with the elements to add

    Returns:
        tuple[T, ...]: A new tuple with all elements from tpl_1 and tpl_2

    Raises:
        TypeError: if tpl_1 or tpl_2 is None
    """
    return require_non_none(tpl_1) + require_non_none(tpl_2)



"""types module tests"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Union, Awaitable

from dev4py.utils.types import Function, T, R, Predicate, Consumer, Supplier, Runnable, BiFunction, U, SyncOrAsync, \
    BiConsumer


class TestFunction:
    """Function type tests"""

    class TestNominalCase:

        def test_empty_function__should__be_equivalent_to_callable_t_r(self):
            """Empty Function should be equivalent to Callable[[T], R]"""
            # GIVEN
            current_type = Function

            # WHEN / THEN
            assert current_type == Callable[[T], R]

        def test_int_to_str_function__should__be_equivalent_to_callable_int_str(self):
            """Function[int, str] should be equivalent to Callable[[int], str]"""
            # GIVEN
            current_type = Function[int, str]

            # WHEN / THEN
            assert current_type == Callable[[int], str]

        def test_int_to_str_function__should__not_be_equivalent_to_callable_str_str(self):
            """Function[int, str] should NOT be equivalent to Callable[[str], str]"""
            # GIVEN
            current_type = Function[int, str]

            # WHEN / THEN
            assert current_type != Callable[[str], str]


class TestPredicate:
    """Predicate type tests"""

    class TestNominalCase:
        def test_empty_predicate__should__be_equivalent_to_callable_t_bool(self):
            """Empty Predicate should be equivalent to Callable[[T], bool]"""
            # GIVEN
            current_type = Predicate

            # WHEN / THEN
            assert current_type == Callable[[T], bool]

        def test_int_predicate__should__be_equivalent_to_callable_int_bool(self):
            """Predicate[int] should be equivalent to Callable[[int], bool]"""
            # GIVEN
            current_type = Predicate[int]

            # WHEN / THEN
            assert current_type == Callable[[int], bool]

        def test_int_predicate__should__not_be_equivalent_to_callable_str_bool(self):
            """Predicate[int] should NOT be equivalent to Callable[[str], bool]"""
            # GIVEN
            current_type = Predicate[int]

            # WHEN / THEN
            assert current_type != Callable[[str], bool]


class TestConsumer:
    """Consumer type tests"""

    class TestNominalCase:
        def test_empty_consumer__should__be_equivalent_to_callable_t_none(self):
            """Empty Consumer should be equivalent to Callable[[T], None]"""
            # GIVEN
            current_type = Consumer

            # WHEN / THEN
            assert current_type == Callable[[T], None]

        def test_int_consumer__should__be_equivalent_to_callable_int_none(self):
            """Consumer[int] should be equivalent to Callable[[int], None]"""
            # GIVEN
            current_type = Consumer[int]

            # WHEN / THEN
            assert current_type == Callable[[int], None]

        def test_int_consumer__should__not_be_equivalent_to_callable_str_none(self):
            """Consumer[int] should NOT be equivalent to Callable[[str], None]"""
            # GIVEN
            current_type = Consumer[int]

            # WHEN / THEN
            assert current_type != Callable[[str], None]


class TestSupplier:
    """Supplier type tests"""

    class TestNominalCase:
        def test_empty_supplier__should__be_equivalent_to_callable_empty_r(self):
            """Empty Supplier should be equivalent to Callable[[], R]"""
            # GIVEN
            current_type = Supplier

            # WHEN / THEN
            assert current_type == Callable[[], R]

        def test_int_supplier__should__be_equivalent_to_callable_int_none(self):
            """Supplier[int] should be equivalent to Callable[[], int]"""
            # GIVEN
            current_type = Supplier[int]

            # WHEN / THEN
            assert current_type == Callable[[], int]

        def test_int_supplier__should__not_be_equivalent_to_callable_str_int(self):
            """Supplier[int] should NOT be equivalent to Callable[[str], R]"""
            # GIVEN
            current_type = Supplier[int]

            # WHEN / THEN
            assert current_type != Callable[[str], int]

        def test_int_supplier__should__not_be_str_supplier(self):
            """Supplier[int] should NOT be equivalent to Callable[[], str]"""
            # GIVEN
            current_type = Supplier[int]

            # WHEN / THEN
            assert current_type != Callable[[], str]


class TestRunnable:
    """Runnable type tests"""

    class TestNominalCase:
        def test_runnable__should__be_equivalent_to_callable_empty_none(self):
            """Empty Supplier should be equivalent to Callable[[], None]"""
            # GIVEN
            current_type = Runnable

            # WHEN / THEN
            assert current_type == Callable[[], None]


class TestBiFunction:
    """BiFunction type tests"""

    class TestNominalCase:

        def test_empty_bifunction__should__be_equivalent_to_callable_t_u_r(self):
            """Empty BiFunction should be equivalent to Callable[[T, U], R]"""
            # GIVEN
            current_type = BiFunction

            # WHEN / THEN
            assert current_type == Callable[[T, U], R]

        def test_int_bool_to_str_bifunction__should__be_equivalent_to_callable_int_bool_str(self):
            """BiFunction[int, bool, str] should be equivalent to Callable[[int, bool], str]"""
            # GIVEN
            current_type = BiFunction[int, bool, str]

            # WHEN / THEN
            assert current_type == Callable[[int, bool], str]

        def test_int_bool_to_str_bifunction__should__not_be_equivalent_to_callable_str_bool_str(self):
            """BiFunction[int, str] should NOT be equivalent to Callable[[str, bool], str]"""
            # GIVEN
            current_type = BiFunction[int, bool, str]

            # WHEN / THEN
            assert current_type != Callable[[str, bool], str]

        def test_int_bool_to_str_bifunction__should__not_be_equivalent_to_callable_int_str_str(self):
            """BiFunction[int, str] should NOT be equivalent to Callable[[int, str], str]"""
            # GIVEN
            current_type = BiFunction[int, bool, str]

            # WHEN / THEN
            assert current_type != Callable[[int, str], str]


class TestSyncOrAsync:
    """SyncOrAsync type tests"""

    class TestNominalCase:
        def test_empty_syncorasync__should__be_equivalent_to_union_awaitable_t_t(self):
            """Empty SyncOrAsync should be equivalent to Union[Awaitable[T], T]"""
            # GIVEN
            current_type = SyncOrAsync

            # WHEN / THEN
            assert current_type == Union[Awaitable[T], T]

        def test_int_syncorasync__should__be_equivalent_to_union_awaitable_int_int(self):
            """SyncOrAsync[int] should be equivalent to Union[Awaitable[int], int]"""
            # GIVEN
            current_type = SyncOrAsync[int]

            # WHEN / THEN
            assert current_type == Union[Awaitable[int], int]

        def test_int_syncorasync__should__not_be_equivalent_to_awaitable_str_int(self):
            """SyncOrAsync[int] should NOT be equivalent to Union[Awaitable[str], int]"""
            # GIVEN
            current_type = SyncOrAsync[int]

            # WHEN / THEN
            assert current_type != Union[Awaitable[str], int]

        def test_int_syncorasync__should__not_be_equivalent_to_awaitable_int_str(self):
            """SyncOrAsync[int] should NOT be equivalent to Union[Awaitable[int], str]"""
            # GIVEN
            current_type = SyncOrAsync[int]

            # WHEN / THEN
            assert current_type != Union[Awaitable[int], str]

        def test_int_syncorasync__should__not_be_equivalent_to_awaitable_str_str(self):
            """SyncOrAsync[int] should NOT be equivalent to Union[Awaitable[str], str]"""
            # GIVEN
            current_type = SyncOrAsync[int]

            # WHEN / THEN
            assert current_type != Union[Awaitable[str], str]


class TestBiConsumer:
    """BiConsumer type tests"""

    class TestNominalCase:

        def test_empty_biconsumer__should__be_equivalent_to_callable_t_u_r(self):
            """Empty BiConsumer should be equivalent to Callable[[T, U], None]"""
            # GIVEN
            current_type = BiConsumer

            # WHEN / THEN
            assert current_type == Callable[[T, U], None]

        def test_int_bool_biconsumer__should__be_equivalent_to_callable_int_bool_none(self):
            """BiConsumer[int, bool] should be equivalent to Callable[[int, bool], None]"""
            # GIVEN
            current_type = BiConsumer[int, bool]

            # WHEN / THEN
            assert current_type == Callable[[int, bool], None]

        def test_int_bool_biconsumer__should__not_be_equivalent_to_callable_str_bool_none(self):
            """BiConsumer[int, str] should NOT be equivalent to Callable[[str, bool], none]"""
            # GIVEN
            current_type = BiConsumer[int, bool]

            # WHEN / THEN
            assert current_type != Callable[[str, bool], None]



"""
The `awaitables` module provides a set of utility functions to simplify Awaitable operations
"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Awaitable, Any, cast

from dev4py.utils import objects
from dev4py.utils.types import T, R, Function, SyncOrAsync


def is_awaitable(value: Any) -> bool:
    """
    If the given value is an Awaitable (Coroutine, Task or Future), returns true, otherwise false

    Returns:
        bool: true if the given value is an Awaitable, otherwise false
    """
    return isinstance(value, Awaitable)


def to_sync_or_async_param_function(
        function: Function[T, SyncOrAsync[R]]
) -> Function[SyncOrAsync[T], Awaitable[R]]:
    """
    If the given Function uses sync parameter, create a new equivalent async function supporting sync or Awaitable
    parameter

    Returned Function specificities:
        1. Is an async function (it means if the given function result is R the new function result is an Awaitable[R])
        2. If the given function is async (i.e. result is an Awaitable[R]), the new function result stay an Awaitable[R]
        3. !WARNING! If the given function expected parameter is an Awaitable, it will be awaited by the wrapper before
        being passed as given function parameter. So, it will probably raise an error depending on your code

    Note: This function can be useful associated with the map function of a `JOptional` with Awaitable value

    Args:
        function: The function with sync parameter

    Returns:
        sync_or_async_param_function: An async function with awaitable parameter

    Raises:
        TypeError: Raises a TypeError if function is None
        Exception: If the given function expected parameter is an Awaitable, it will be awaited by the wrapper before
        being passed as function parameter, and, it will probably raise an error depending on function your code
    """
    objects.require_non_none(function)

    async def sync_or_async_param_function(param: SyncOrAsync[T]) -> R:
        result: SyncOrAsync[R] = function((await cast(Awaitable[T], param)) if is_awaitable(param) else cast(T, param))
        return await cast(Awaitable[R], result) if is_awaitable(result) else cast(R, result)

    return sync_or_async_param_function


def to_awaitable(value: SyncOrAsync[T]) -> Awaitable[T]:
    """
    This function map a SyncOrAsync[T] to an Awaitable[T]

    Note: This can be useful to work with SyncOrAsync values in order to always works with `await`. Without this
    function, you always have to call is_awaitable in order to know if you have to use `await` or not.

    Args:
        value: the value to map

    Returns:
        Awaitable[T]: a value to await
    """

    async def _sync_to_awaitable(param: T) -> T:
        return param

    return cast(Awaitable[T], value) if is_awaitable(value) else _sync_to_awaitable(cast(T, value))


async def async_none() -> None:
    """
    This function provides an Awaitable[None]

    Returns:
        Awaitable[None]: An awaitable of None

    """
    return None



"""Stream module tests"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from concurrent.futures import Executor, ProcessPoolExecutor, ThreadPoolExecutor
from dataclasses import FrozenInstanceError
from functools import partial
from multiprocessing import Manager
from sys import maxsize
from typing import Optional, Iterable, Final, Type, Iterator

from pytest import raises, fixture

from dev4py.utils import ParallelConfiguration, Stream, JOptional, lists
from dev4py.utils.collectors import to_list, Collector
from dev4py.utils.objects import non_none, is_none, to_self
from dev4py.utils.pipeline import StepPipeline, StepResult
from dev4py.utils.types import BiFunction, Function, Predicate, Supplier, BiConsumer, Consumer


class _FakeExecutor(Executor):
    """Fake Executor for tests"""
    ...


class TestParallelConfiguration:
    """ParallelConfiguration class tests"""

    class TestConstructor:
        """Constructor tests"""

        class TestNominalCase:
            def test_existing_parameters__should__create_new_parallel_configuration(self) -> None:
                """When all parameters are set, should create a new ParallelConfiguration with given parameters"""
                # GIVEN
                executor: Executor = _FakeExecutor()
                chunksize: int = 2

                # WHEN
                parallel_config: ParallelConfiguration = ParallelConfiguration(
                    executor=executor, chunksize=chunksize
                )

                # THEN
                assert parallel_config.executor == executor
                assert parallel_config.chunksize == chunksize

            def test_no_chunksize__should__create_new_parallel_configuration_with_chunksize_one(self) -> None:
                """When uses default chunksize, should create a new ParallelConfiguration with chunksize set to 1"""
                # GIVEN
                executor: Executor = _FakeExecutor()
                chunksize: int = 1

                # WHEN
                parallel_config: ParallelConfiguration = ParallelConfiguration(executor=executor)

                # THEN
                assert parallel_config.executor == executor
                assert parallel_config.chunksize == chunksize

        class TestErrorCase:
            def test_none_executor__should_raise_type_error(self) -> None:
                """When executor is None, should raise TypeError exception"""
                # GIVEN
                chunksize: int = 1

                # WHEN / THEN
                with raises(TypeError) as error:
                    # noinspection PyTypeChecker
                    ParallelConfiguration(executor=None, chunksize=chunksize)

                assert str(error.value) == "executor must be non None"

            def test_none_chunksize__should_raise_type_error(self) -> None:
                """When chunksize is None, should raise TypeError exception"""
                # GIVEN
                executor: Executor = _FakeExecutor()

                # WHEN / THEN
                with raises(TypeError) as error:
                    # noinspection PyTypeChecker
                    ParallelConfiguration(executor=executor, chunksize=None)

                assert str(error.value) == "chunksize must be non None"

            def test_less_than_one_chunksize__should_raise_value_error(self) -> None:
                """When chunksize is less than one, should raise ValueError exception"""
                # GIVEN
                executor: Executor = _FakeExecutor()

                # WHEN / THEN
                with raises(ValueError) as error:
                    # noinspection PyTypeChecker
                    ParallelConfiguration(executor=executor, chunksize=0)

                assert str(error.value) == "chunksize must be greater than or equals to 1"

    class TestSetters:
        """Setters tests"""

        class TestErrorCase:
            def test_set_new_executor__should__raise_frozen_instance_error(self) -> None:
                """When set a new executor, should raise a FrozenInstanceError exception"""
                # GIVEN
                executor: Executor = _FakeExecutor()
                chunksize: int = 1
                parallel_config: ParallelConfiguration = ParallelConfiguration(
                    executor=executor, chunksize=chunksize
                )

                # WHEN / THEN
                new_executor: Executor = _FakeExecutor()
                with raises(FrozenInstanceError):
                    # noinspection PyDataclass
                    parallel_config.executor = new_executor

            def test_set_new_chunksize__should__raise_frozen_instance_error(self) -> None:
                """When set a new chunksize, should raise a FrozenInstanceError exception"""
                # GIVEN
                executor: Executor = _FakeExecutor()
                chunksize: int = 1
                parallel_config: ParallelConfiguration = ParallelConfiguration(
                    executor=executor, chunksize=chunksize
                )

                # WHEN / THEN
                new_chunksize: int = 10
                with raises(FrozenInstanceError):
                    # noinspection PyDataclass
                    parallel_config.chunksize = new_chunksize


class TestStream:
    """Stream class tests"""
    CONSTRUCTOR_ERROR_MSG: Final[str] = "Stream private constructor! Please use Stream.of"
    TEST_RANGE_MIN_VALUE = 0
    TEST_RANGE_MAX_VALUE = 99
    TEST_VALUES: range = range(TEST_RANGE_MIN_VALUE, TEST_RANGE_MAX_VALUE + 1)
    TEST_STR_VALUES: list[str] = [str(i) for i in TEST_VALUES]

    @staticmethod
    def _predicate_greater_or_eq(i: int, min_value: int) -> bool:
        """test predicate because can't pickle lambda"""
        return i >= min_value

    @staticmethod
    def _predicate_less_than(i: int, value: int) -> bool:
        """test predicate because can't pickle lambda"""
        return i < value

    @fixture(
        scope="module",
        params=[
            None,
            (ProcessPoolExecutor, 1),
            (ThreadPoolExecutor, 1),
            (ProcessPoolExecutor, 3),
            (ThreadPoolExecutor, 3),
        ]
    )
    def parallel_config_fixture(self, request) -> Optional[ParallelConfiguration]:
        """Fixture to manage test ParallelConfiguration"""
        # INIT
        request_param: tuple[Type[Executor], int] = request.param
        parallel_config: Optional[ParallelConfiguration] = None if is_none(request_param) else \
            ParallelConfiguration(
                executor=request_param[0](),
                chunksize=request_param[1]
            )

        # RETURN
        yield parallel_config

        # TEAR DOWN
        if non_none(parallel_config):
            parallel_config.executor.shutdown()

    @fixture
    def test_stream(self, parallel_config_fixture) -> Stream[int]:
        """Fixture to create a new set of test stream"""
        yield Stream.of(*TestStream.TEST_VALUES).parallel(parallel_config_fixture)

    class TestConstructor:
        """Constructor tests"""

        class TestNominalCase:
            def test_valid_create_key__should__create_a_new_stream(self) -> None:
                """When create key is valid, should create a new Stream"""
                # GIVEN
                values_function: BiFunction[Optional[ParallelConfiguration], bool, Iterable[int]] = \
                    lambda pc, order: lists.empty_list()
                pipeline: StepPipeline[int, int] = StepPipeline.of(lambda i: StepResult(value=i))
                parallel_config: JOptional[ParallelConfiguration] = JOptional.empty()
                ordered_execution: bool = False
                # noinspection PyUnresolvedReferences
                create_key: object = Stream._Stream__CREATE_KEY

                # WHEN
                stream: Stream[int] = Stream(
                    values_function=values_function,
                    pipeline=pipeline,
                    parallel_config=parallel_config,
                    ordered_execution=ordered_execution,
                    create_key=create_key
                )

                # THEN
                assert stream._values_function == values_function
                assert stream._pipeline == pipeline
                assert stream._parallel_config == parallel_config
                assert stream._ordered_execution == ordered_execution

        class TestErrorCase:
            def test_none_create_key__should__raise_assertion_error(self) -> None:
                """When create key is none, should raise AssertionError with private constructor message"""
                # GIVEN
                values_function: BiFunction[Optional[ParallelConfiguration], bool, Iterable[int]] = \
                    lambda pc, order: lists.empty_list()
                pipeline: StepPipeline[int, int] = StepPipeline.of(lambda i: StepResult(value=i))
                parallel_config: JOptional[ParallelConfiguration] = JOptional.empty()
                ordered_execution: bool = False

                # WHEN / THEN
                with raises(AssertionError) as error:
                    Stream(
                        values_function=values_function,
                        pipeline=pipeline,
                        parallel_config=parallel_config,
                        ordered_execution=ordered_execution,
                        create_key=None
                    )

                assert str(error.value) == TestStream.CONSTRUCTOR_ERROR_MSG

            def test_invalid_create_key__should__raise_assertion_error(self) -> None:
                """When create key is not valid, should raise AssertionError with private constructor message"""
                # GIVEN
                values_function: BiFunction[Optional[ParallelConfiguration], bool, Iterable[int]] = \
                    lambda pc, order: lists.empty_list()
                pipeline: StepPipeline[int, int] = StepPipeline.of(lambda i: StepResult(value=i))
                parallel_config: JOptional[ParallelConfiguration] = JOptional.empty()
                ordered_execution: bool = False
                create_key: object = object()

                # WHEN / THEN
                with raises(AssertionError) as error:
                    Stream(
                        values_function=values_function,
                        pipeline=pipeline,
                        parallel_config=parallel_config,
                        ordered_execution=ordered_execution,
                        create_key=create_key
                    )

                assert str(error.value) == TestStream.CONSTRUCTOR_ERROR_MSG

            def test_none_values_function__should__raise_type_error(self) -> None:
                """When values_function is None, should raise TypeError exception"""
                # GIVEN
                pipeline: StepPipeline[int, int] = StepPipeline.of(lambda i: StepResult(value=i))
                parallel_config: JOptional[ParallelConfiguration] = JOptional.empty()
                ordered_execution: bool = False
                # noinspection PyUnresolvedReferences
                create_key: object = Stream._Stream__CREATE_KEY

                # WHEN / THEN
                with raises(TypeError):
                    Stream(
                        values_function=None,
                        pipeline=pipeline,
                        parallel_config=parallel_config,
                        ordered_execution=ordered_execution,
                        create_key=create_key
                    )

            def test_none_pipeline__should__raise_type_error(self) -> None:
                """When pipeline is None, should raise TypeError exception"""
                # GIVEN
                values_function: BiFunction[Optional[ParallelConfiguration], bool, Iterable[int]] = \
                    lambda pc, order: lists.empty_list()
                parallel_config: JOptional[ParallelConfiguration] = JOptional.empty()
                ordered_execution: bool = False
                # noinspection PyUnresolvedReferences
                create_key: object = Stream._Stream__CREATE_KEY

                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    Stream(
                        values_function=values_function,
                        pipeline=None,
                        parallel_config=parallel_config,
                        ordered_execution=ordered_execution,
                        create_key=create_key
                    )

            def test_none_parallel_config__should__raise_type_error(self) -> None:
                """When parallel_config is None, should raise TypeError exception"""
                # GIVEN
                values_function: BiFunction[Optional[ParallelConfiguration], bool, Iterable[int]] = \
                    lambda pc, order: lists.empty_list()
                pipeline: StepPipeline[int, int] = StepPipeline.of(lambda i: StepResult(value=i))
                ordered_execution: bool = False
                # noinspection PyUnresolvedReferences
                create_key: object = Stream._Stream__CREATE_KEY

                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    Stream(
                        values_function=values_function,
                        pipeline=pipeline,
                        parallel_config=None,
                        ordered_execution=ordered_execution,
                        create_key=create_key
                    )

            def test_none_ordered_execution__should__raise_type_error(self) -> None:
                """When ordered_execution is None, should raise TypeError exception"""
                # GIVEN
                values_function: BiFunction[Optional[ParallelConfiguration], bool, Iterable[int]] = \
                    lambda pc, order: lists.empty_list()
                pipeline: StepPipeline[int, int] = StepPipeline.of(lambda i: StepResult(value=i))
                parallel_config: JOptional[ParallelConfiguration] = JOptional.empty()
                # noinspection PyUnresolvedReferences
                create_key: object = Stream._Stream__CREATE_KEY

                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    Stream(
                        values_function=values_function,
                        pipeline=pipeline,
                        parallel_config=parallel_config,
                        ordered_execution=None,
                        create_key=create_key
                    )

    class TestEmpty:
        """empty class method tests"""

        class TestNominalCase:
            def test_should__return_an_empty_stream(self) -> None:
                """Should return an empty stream"""
                # GIVEN / WHEN
                stream: Stream[int] = Stream.empty()

                # THEN:
                assert stream.count() == 0

    class TestOf:
        """of class method tests"""

        class TestNominalCase:

            def test_none__should__return_a_stream_of_none(self) -> None:
                """When None value should return a stream of None"""
                # GIVEN / WHEN
                stream: Stream[None] = Stream.of(None)

                # THEN
                assert stream.to_generator().__next__() is None
                assert stream._ordered_execution is False

            def test_existing_parameters__should__return_a_stream_of_parameters(self) -> None:
                """When parameters exists should return a stream of given parameters"""
                # GIVEN / WHEN
                stream: Stream[int] = Stream.of(1, 2, 3, 4)

                # THEN
                assert stream.to_list() == [1, 2, 3, 4]
                assert stream._ordered_execution is False

    class TestOfIterable:
        """of_iterable class method tests"""

        class TestNominalCase:
            def test_iterable_exists__should__return_a_stream_of_iterable_values(self) -> None:
                """When iterable is provided, should return a stream of given iterable values"""
                # GIVEN / WHEN
                stream: Stream[int] = Stream.of_iterable([1, 2, 3, 4])

                # THEN
                assert stream.to_list() == [1, 2, 3, 4]

        class TestErrorCase:
            def test_none_iterable__should__raise_type_error(self) -> None:
                """When iterable is None, should raise TypeError exception"""
                # GIVEN / WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    Stream.of_iterable(None)

            def test_not_iterable__should__raise_type_error_on_terminal_operation(self) -> None:
                """When iterable is not iterable, should raise TypeError exception on terminal operation"""
                # GIVEN / WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    Stream.of_iterable(1).to_list()

    class TestMap:
        """map method tests"""

        @staticmethod
        def _mapper(i: int) -> str:
            """map mapper test because can't pickle lambda"""
            return str(i)

        class TestNominalCase:
            def test_mapper_exists__should__return_a_stream_with_given_mapper(self, test_stream: Stream[int]) -> None:
                """When mapper exists, should return a stream that uses this mapper on terminal operation"""
                # GIVEN test_stream
                mapper: Function[int, str] = TestStream.TestMap._mapper

                # WHEN
                new_stream: Stream[str] = test_stream.map(mapper)

                # THEN
                assert all(elem in TestStream.TEST_STR_VALUES for elem in new_stream.to_list())

        class TestErrorCase:
            def test_none_mapper_should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When mapper is None, should raise TypeError exception"""
                # GIVEN test_stream
                # WHEN / THEN
                with raises(TypeError):
                    test_stream.map(None)

    class TestParallel:
        """parallel method tests"""

        class TestNominalCase:
            def test_parallel_config_exists__should__return_a_stream_with_parallel_configuration(self) -> None:
                """When parallel_config is provided, should return a stream with the parallel_config"""
                # GIVEN
                executor: Executor = _FakeExecutor()
                chunksize: int = 3
                parallel_config: ParallelConfiguration = ParallelConfiguration(executor=executor, chunksize=chunksize)

                # WHEN
                new_stream: Stream[int] = Stream.of_iterable(TestStream.TEST_VALUES).parallel(parallel_config)

                # THEN
                assert new_stream.is_parallel()
                assert new_stream._parallel_config.filter(lambda pc: pc == parallel_config).is_present()

            def test_none_parallel_config__should__return_a_sequential_stream(self) -> None:
                # GIVEN / WHEN
                new_stream: Stream[int] = Stream.of_iterable(TestStream.TEST_VALUES).parallel(None)

                # THEN
                assert new_stream.is_parallel() is False

    class TestIsParallel:
        """is_parallel method tests"""

        class TestNominalCase:
            def test_parallel_stream__should__return_true(self) -> None:
                """When stream uses parallel configuration, should return True"""
                # GIVEN
                executor: Executor = _FakeExecutor()
                chunksize: int = 3
                parallel_config: ParallelConfiguration = ParallelConfiguration(executor=executor, chunksize=chunksize)
                stream: Stream[int] = Stream.of_iterable(TestStream.TEST_VALUES).parallel(parallel_config)

                # WHEN
                result: bool = stream.is_parallel()

                # THEN
                assert result is True

            def test_sequential_stream__should__return_false(self) -> None:
                """When stream is sequential, should return False"""

                # GIVEN
                stream: Stream[int] = Stream.of_iterable(TestStream.TEST_VALUES)

                # WHEN
                result: bool = stream.is_parallel()

                # THEN
                assert result is False

    class TestSequential:
        """sequential method tests"""

        class TestNominalCase:

            def test_sequential_stream__should__return_self_and_stay_sequential(self) -> None:
                """When stream is sequential, should return self"""
                # GIVEN
                stream: Stream[int] = Stream.empty().parallel(None)

                # WHEN
                new_stream: Stream[int] = stream.sequential()

                # THEN
                assert new_stream == stream
                assert new_stream.is_parallel() is False

            def test_parallel_stream__should__return_self_as_sequential_stream(self) -> None:
                """When stream is parallel, should return self with sequential configuration"""
                # GIVEN
                stream: Stream[int] = Stream.empty().parallel(ParallelConfiguration(_FakeExecutor()))

                # WHEN
                new_stream: Stream[int] = stream.sequential()

                # THEN
                assert new_stream == stream
                assert new_stream.is_parallel() is False

    class TestFilter:
        """filter method tests"""
        _MIN_VALUE: int = 20

        class TestNominalCase:
            def test_useful_filter__should__remove_values_from_the_stream(self, test_stream: Stream[int]) -> None:
                """
                When some values are concerned by the filter, should return a stream without these values on terminal
                operation
                """
                # GIVEN test_stream
                min_value: int = TestStream.TestFilter._MIN_VALUE
                test_filter: Predicate[int] = partial(TestStream._predicate_greater_or_eq, min_value=min_value)

                # WHEN
                result: list[int] = test_stream.filter(test_filter).to_list()

                # THEN
                assert len(result) == len(TestStream.TEST_VALUES) - min_value
                for i in result:
                    assert test_filter(i) is True
                    assert i in TestStream.TEST_VALUES
                assert all(elem not in result for elem in range(0, min_value))

            def test_useless_filter__should__return_the_stream_with_all_values(self, test_stream: Stream[int]) -> None:
                """
                When no value is concerned by the filter, should return a stream with all values on terminal operation
                """
                # GIVEN test_stream
                min_value: int = -(maxsize - 1)
                test_filter: Predicate[int] = partial(TestStream._predicate_greater_or_eq, min_value=min_value)

                # WHEN
                result: list[int] = test_stream.filter(test_filter).to_list()

                # THEN
                assert len(result) == len(TestStream.TEST_VALUES)
                assert all(elem in result for elem in TestStream.TEST_VALUES)

        class TestErrorCase:
            def test_none_filter__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When filter is None, should raise TypeError exception"""
                # Given test_stream
                # WHEN / THEN
                with raises(TypeError):
                    test_stream.filter(None)

    class TestCollect:
        """collect method test"""

        class TestNominalCase:
            def test_collector_exists__should__return_collected_values(self, test_stream: Stream[int]) -> None:
                """When collector exists, should return values as described by collector"""
                # GIVEN test_stream
                collector: Collector[int, list[int]] = to_list()

                # WHEN
                result: list[int] = test_stream.collect(collector)

                # THEN
                assert isinstance(result, list)
                assert len(result) == len(TestStream.TEST_VALUES)
                assert all(elem in result for elem in TestStream.TEST_VALUES)

        class TestErrorCase:
            def test_none_collector__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When collector is None, should raise TypeError exception"""
                # GIVEN test_stream
                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    test_stream.collect(None)

    class TestCollectFrom:
        """collect_from method test"""

        class TestNominalCase:
            def test_existing_parameters__should__return_collected_values(self, test_stream: Stream[int]) -> None:
                """When all parameters are given, should return values as described by these parameters"""
                # GIVEN test_stream
                supplier: Supplier[list[int]] = lists.empty_list
                accumulator: BiConsumer[list[int], int] = lists.append
                combiner: BiConsumer[list[int], list[int]] = lists.extend

                # WHEN
                result: list[int] = test_stream.collect_from(
                    supplier=supplier, accumulator=accumulator, combiner=combiner
                )

                # THEN
                assert isinstance(result, list)
                assert len(result) == len(TestStream.TEST_VALUES)
                assert all(elem in result for elem in TestStream.TEST_VALUES)

        class TestErrorCase:
            def test_none_supplier__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When supplier is None, should raise TypeError exception"""
                # GIVEN test_stream
                accumulator: BiConsumer[list[int], int] = lists.append
                combiner: BiConsumer[list[int], list[int]] = lists.extend

                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    test_stream.collect_from(supplier=None, accumulator=accumulator, combiner=combiner)

            def test_none_accumulator__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When accumulator is None, should raise TypeError exception"""
                # GIVEN test_stream
                supplier: Supplier[list[int]] = lists.empty_list
                combiner: BiConsumer[list[int], list[int]] = lists.extend

                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    test_stream.collect_from(supplier=supplier, accumulator=None, combiner=combiner)

            def test_none_combiner__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When combiner is None, should raise TypeError exception"""
                # GIVEN test_stream
                supplier: Supplier[list[int]] = lists.empty_list
                accumulator: BiConsumer[list[int], int] = lists.append

                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    test_stream.collect_from(supplier=supplier, accumulator=accumulator, combiner=None)

    class TestToList:
        """to_list method tests"""

        class TestNominalCase:
            def test_should__return_stream_values_as_list(self, test_stream: Stream[int]) -> None:
                """Should return stream values as list"""
                # GIVEN test_stream
                # WHEN
                result: list[int] = test_stream.to_list()

                # THEN
                assert isinstance(result, list)
                assert len(result) == len(TestStream.TEST_VALUES)
                assert all(elem in result for elem in TestStream.TEST_VALUES)

    class TestToTuple:
        """to_tuple method tests"""

        class TestNominalCase:
            def test_should__return_stream_values_as_tuple(self, test_stream: Stream[int]) -> None:
                """Should return stream values as tuple"""
                # GIVEN test_stream
                # WHEN
                result: tuple[int, ...] = test_stream.to_tuple()

                # THEN
                assert isinstance(result, tuple)
                assert len(result) == len(TestStream.TEST_VALUES)
                assert all(elem in result for elem in TestStream.TEST_VALUES)

    class TestToDict:
        """to_dict method tests"""

        class TestNominalCase:
            def test_existing_mappers_should__return_stream_values_as_dict(self, test_stream: Stream[int]) -> None:
                """When key/value mappers are provided, should return a dict as described by mappers"""
                # GIVEN test_stream
                key_mapper: Function[int, str] = str
                value_mapper: Function[int, int] = to_self

                # WHEN
                result: dict[str, int] = test_stream.to_dict(key_mapper, value_mapper)

                # THEN
                assert isinstance(result, dict)
                assert len(result) == len(TestStream.TEST_VALUES)
                assert all(elem in result.values() for elem in TestStream.TEST_VALUES)
                assert all(elem in result.keys() for elem in TestStream.TEST_STR_VALUES)

    class TestErrorCase:
        def test_none_key_mapper__should__raise_type_error(self, test_stream: Stream[int]) -> None:
            """When key mapper is None should raise a TypeError exception"""
            # GIVEN test_stream
            value_mapper: Function[int, int] = to_self

            # WHEN / THEN
            with raises(TypeError):
                test_stream.to_dict(None, value_mapper)

        def test_none_value_mapper__should__raise_type_error(self, test_stream: Stream[int]) -> None:
            """When value mapper is None should raise a TypeError exception"""
            # GIVEN test_stream
            key_mapper: Function[int, str] = str

            # WHEN / THEN
            with raises(TypeError):
                test_stream.to_dict(key_mapper, None)

    class TestAllMatch:
        """all_match method tests"""

        class TestNominalCase:
            def test_all_match_given_predicate_should__return_true(self, test_stream: Stream[int]) -> None:
                """When all values match with the given predicate, should return True"""
                # GIVEN test_stream
                predicate: Predicate[int] = non_none

                # WHEN
                result: bool = test_stream.all_match(predicate)

                # THEN
                assert result is True

            def test_at_least_one_value_not_match_given_predicate_should__return_false(
                    self, test_stream: Stream[int]
            ) -> None:
                """When at least one value doesn't match with the given predicate, should return False"""
                # GIVEN test_stream
                predicate: Predicate[int] = partial(TestStream._predicate_greater_or_eq, min_value=10)

                # WHEN
                result: bool = test_stream.all_match(predicate)

                # THEN
                assert result is False

            def test_empty_stream__should__return_true(self) -> None:
                """When stream is empty, should return True"""
                # GIVEN
                stream: Stream[int] = Stream.empty()
                predicate: Predicate[int] = partial(TestStream._predicate_greater_or_eq, min_value=10)

                # WHEN
                result: bool = stream.all_match(predicate)

                # THEN
                assert result is True

        class TestErrorCase:
            def test_none_predicate__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When predicate is None should raise a TypeError exception"""
                # GIVEN test_stream
                # WHEN / THEN
                with raises(TypeError):
                    test_stream.all_match(None)

    class TestAnyMatch:
        """any_match method tests"""

        class TestNominalCase:
            def test_all_match_given_predicate_should__return_true(self, test_stream: Stream[int]) -> None:
                """When all values match with the given predicate, should return True"""
                # GIVEN test_stream
                predicate: Predicate[int] = non_none

                # WHEN
                result: bool = test_stream.any_match(predicate)

                # THEN
                assert result is True

            def test_one_value_match_given_predicate_should__return_true(self, test_stream: Stream[int]) -> None:
                """When all values match with the given predicate, should return True"""
                # GIVEN test_stream
                predicate: Predicate[int] = \
                    partial(TestStream._predicate_greater_or_eq, min_value=TestStream.TEST_RANGE_MAX_VALUE)

                # WHEN
                result: bool = test_stream.any_match(predicate)

                # THEN
                assert result is True

            def test_all_not_match_given_predicate_should__return_false(
                    self, test_stream: Stream[int]
            ) -> None:
                """When at least one value doesn't match with the given predicate, should return False"""
                # GIVEN test_stream
                predicate: Predicate[int] = \
                    partial(TestStream._predicate_greater_or_eq, min_value=TestStream.TEST_RANGE_MAX_VALUE + 1)

                # WHEN
                result: bool = test_stream.any_match(predicate)

                # THEN
                assert result is False

            def test_empty_stream__should__return_false(self) -> None:
                """When stream is empty, should return False"""
                # GIVEN
                stream: Stream[int] = Stream.empty()
                predicate: Predicate[int] = partial(TestStream._predicate_greater_or_eq, min_value=10)

                # WHEN
                result: bool = stream.any_match(predicate)

                # THEN
                assert result is False

        class TestErrorCase:
            def test_none_predicate__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When predicate is None should raise a TypeError exception"""
                # GIVEN test_stream
                # WHEN / THEN
                with raises(TypeError):
                    test_stream.any_match(None)

    class TestCount:
        """count method tests"""

        class TestNominalCase:
            def test_should__return_the_count_of_stream_elements(self, test_stream: Stream[int]) -> None:
                """Should return the count of elements in this stream"""
                # GIVEN test_stream
                # WHEN
                result: int = test_stream.count()

                # THEN
                assert isinstance(result, int)
                assert result == len(TestStream.TEST_VALUES)

    class TestForEach:
        """for_each method tests"""

        class TestNominalCase:

            def test_consumer_exists_should__call_consumer_for_each_stream_element(
                    self, test_stream: Stream[int]
            ) -> None:
                with Manager() as manager:
                    # GIVEN test_stream
                    result_list: list[int] = manager.list()
                    consumer: Consumer[int] = result_list.append

                    # WHEN
                    test_stream.for_each(consumer)

                    # THEN
                    assert len(result_list) == len(TestStream.TEST_VALUES)
                    assert all(elem in result_list for elem in TestStream.TEST_VALUES)

        class TestErrorCase:
            def test_none_consumer__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When consumer is None should raise a TypeError exception"""
                # GIVEN test_stream
                # WHEN / THEN
                with raises(TypeError):
                    test_stream.for_each(None)

    class TestFlatMap:
        """flat_map method tests"""

        @staticmethod
        def _mapper(i: int):
            """flat_map test mapper because can't pickle lambda"""
            # lambda i: Stream.of(i).map(str)
            return Stream.of(i).map(str)

        class TestNominalCase:
            def test_mapper_exists__should__return_stream_with_mapper_each_element_by_mapper_stream_result(
                    self, test_stream: Stream[int]
            ) -> None:
                """
                When mapper exists, should return a stream consisting of the results of replacing each element of this
                stream with the contents of a mapped stream produced by applying the provided mapping function to each
                element
                """
                # GIVEN test_stream
                mapper: Function[int, Stream[str]] = TestStream.TestFlatMap._mapper

                # WHEN
                new_stream: Stream[str] = test_stream.flat_map(mapper)

                # THEN
                result: list[str] = new_stream.to_list()
                assert len(result) == len(TestStream.TEST_STR_VALUES)
                assert all(elem in result for elem in TestStream.TEST_STR_VALUES)

        class TestErrorCase:
            def test_none_mapper__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When mapper is None, should raise TypeError exception"""
                # GIVEN test_stream
                # WHEN / THEN
                with raises(TypeError):
                    test_stream.flat_map(None)

    class TestOrderedExecution:
        """ordered_execution method tests"""

        class TestNominalCase:
            def test_ordered_is_true__should__return_mapped_element_in_the_source_order(
                    self, test_stream: Stream[int]
            ) -> None:
                """When ordered is True, should return elements in the source order on terminal operation"""
                # GIVEN test_stream
                ordered_execution: bool = True

                # WHEN
                new_stream: Stream[str] = test_stream.map(str).ordered_execution(ordered_execution)

                # THEN
                result: list[str] = new_stream.to_list()
                assert new_stream._ordered_execution == ordered_execution
                assert result == TestStream.TEST_STR_VALUES

            def test_ordered_is_false__should__return_mapped_source_element_in_random_order(
                    self, test_stream: Stream[int]
            ) -> None:
                """
                When ordered is False, should return all source elements on terminal operation (but not necessary in the
                same order)
                """
                # GIVEN test_stream
                ordered_execution: bool = False

                # WHEN
                new_stream: Stream[str] = test_stream.map(str).ordered_execution(ordered_execution)

                # THEN
                result: list[str] = new_stream.to_list()
                assert new_stream._ordered_execution == ordered_execution
                assert len(result) == len(TestStream.TEST_STR_VALUES)
                assert all(elem in result for elem in TestStream.TEST_STR_VALUES)

        class TestErrorCase:
            def test_none_parameter__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When parameter is None, should raise TypeError exception"""
                # GIVEN test_stream
                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    test_stream.ordered_execution(None)

    class TestUnordered:
        """unordered method tests"""

        class TestNominalCase:
            def test_should__return_mapped_source_element_in_random_order(
                    self, test_stream: Stream[int]
            ) -> None:
                """
                Should return all source elements on terminal operation (but not necessary in the same order)
                """
                # GIVEN test_stream
                # WHEN
                new_stream: Stream[str] = test_stream.map(str).unordered()

                # THEN
                result: list[str] = new_stream.to_list()
                assert new_stream._ordered_execution == False
                assert len(result) == len(TestStream.TEST_STR_VALUES)
                assert all(elem in result for elem in TestStream.TEST_STR_VALUES)

    class TestFindFirst:
        """find_first method tests"""

        class TestNominalCase:

            def test_ordered_stream__should__return_first_encountered_element_in_the_source_order(
                    self, test_stream: Stream[int]
            ) -> None:
                """When stream is ordered, should return the first mapped element in the source order"""
                # GIVEN
                stream: Stream[str] = test_stream.ordered_execution().map(str)

                # WHEN
                result: JOptional[str] = stream.find_first()

                # THEN
                assert result.is_present() is True
                assert result.get() == TestStream.TEST_STR_VALUES[0]

            def test_unordered_stream__should__return_first_encountered_element_in_the_source_order(
                    self, test_stream: Stream[int]
            ) -> None:
                """When stream is ordered, should return the first mapped element in the source order"""
                # GIVEN
                stream: Stream[str] = test_stream.unordered().map(str)

                # WHEN
                result: JOptional[str] = stream.find_first()

                # THEN
                assert result.is_present() is True
                assert result.get() == TestStream.TEST_STR_VALUES[0]

            def test_empty_stream__should__return_empty_joptional(self) -> None:
                """When stream is empty, should return an empty JOptional"""
                # GIVEN
                stream: Stream[str] = Stream.empty().map(str)

                # WHEN
                result: JOptional[str] = stream.find_first()

                # THEN
                assert result.is_present() is False

    class TestFindAny:
        """find_any method tests"""

        class TestNominalCase:

            def test_ordered_stream__should__return_first_encountered_element_in_the_source_order(
                    self, test_stream: Stream[int]
            ) -> None:
                """When stream is ordered, should return the first mapped element in the source order"""
                # GIVEN
                stream: Stream[str] = test_stream.ordered_execution().map(str)

                # WHEN
                result: JOptional[str] = stream.find_any()

                # THEN
                assert result.is_present() is True
                assert result.get() == TestStream.TEST_STR_VALUES[0]

            def test_unordered_stream__should__return_any_mapped_element(
                    self, test_stream: Stream[int]
            ) -> None:
                """When stream is ordered, should return the first mapped element in the source order"""
                # GIVEN
                stream: Stream[str] = test_stream.unordered().map(str)

                # WHEN
                result: JOptional[str] = stream.find_any()

                # THEN
                assert result.is_present() is True
                assert result.get() in TestStream.TEST_STR_VALUES

            def test_empty_stream__should__return_empty_joptional(self) -> None:
                """When stream is empty, should return an empty JOptional"""
                # GIVEN
                stream: Stream[str] = Stream.empty().map(str)

                # WHEN
                result: JOptional[str] = stream.find_any()

                # THEN
                assert result.is_present() is False

    class TestDistinct:
        """distinct method tests"""

        class TestNominalCase:
            def test_empty_stream__should__return_an_empty_stream(self) -> None:
                """When stream is empty, should return an empty stream"""
                # GIVEN
                stream: Stream[int] = Stream.empty()

                # WHEN
                new_stream: Stream[int] = stream.distinct()

                # THEN
                assert new_stream.to_list() == lists.empty_list()

            def test_stream_with_values__should__remove_duplicated_values(
                    self, parallel_config_fixture: ParallelConfiguration
            ) -> None:
                """When values exist, should remove duplicated values"""
                # GIVEN
                values: list[int] = [1, 2, 1, 3, 3, 4, 5, 4]
                stream: Stream[int] = Stream.of_iterable(values).parallel(parallel_config_fixture)

                # WHEN
                new_stream: Stream[int] = stream.distinct()

                # THEN
                expected_values: list[int] = [1, 2, 3, 4, 5]
                result: list[int] = new_stream.to_list()
                assert len(result) == len(expected_values)
                assert all(value in result for value in expected_values)

    class TestDropWhile:
        """drop_while method tests"""

        class TestNominalCase:
            def test_empty_stream__should__return_an_empty_stream(self) -> None:
                """When stream is empty, should return an empty stream"""
                # GIVEN
                stream: Stream[int] = Stream.empty()
                min_value: int = 10
                predicate: Predicate[int] = partial(TestStream._predicate_greater_or_eq, min_value=min_value)

                # WHEN
                new_stream: Stream[int] = stream.drop_while(predicate)

                # THEN
                assert new_stream.to_list() == lists.empty_list()

            def test_ordered_stream_with_matched_predicate__should__return_stream_with_dropped_values(
                    self, test_stream: Stream[int]
            ) -> None:
                """
                When stream is ordered and predicate match at least one value, should return a stream with dropped
                values
                """
                # GIVEN
                stream: Stream[int] = test_stream.ordered_execution()
                min_value: int = 10
                predicate: Predicate[int] = partial(TestStream._predicate_less_than, value=min_value)

                # WHEN
                new_stream: Stream[int] = stream.drop_while(predicate)

                # THEN
                result: list[int] = new_stream.to_list()
                assert len(result) == len(TestStream.TEST_VALUES) - min_value
                assert all(value in result for value in range(min_value, TestStream.TEST_RANGE_MAX_VALUE + 1))

            def test_not_matched_predicate__should__return_stream_with_all_values(
                    self, test_stream: Stream[int]
            ) -> None:
                """When no value match, should return a stream with all values"""
                # GIVEN
                predicate: Predicate[int] = is_none

                # WHEN
                new_stream: Stream[int] = test_stream.drop_while(predicate)

                # THEN
                result: list[int] = new_stream.to_list()
                assert len(result) == len(TestStream.TEST_VALUES)
                assert all(value in result for value in TestStream.TEST_VALUES)

            def test_unordered_stream_with_valid_predicate__should__return_a_stream_with_some_dropped_values(
                    self, test_stream: Stream[int]
            ) -> None:
                """
                When stream is unordered and predicate match at least one value, should return a stream with some
                dropped values
                """
                # GIVEN
                stream: Stream[int] = test_stream.unordered()
                min_value: int = 10
                predicate: Predicate[int] = partial(TestStream._predicate_less_than, value=min_value)

                # WHEN
                new_stream: Stream[int] = stream.drop_while(predicate)

                # THEN
                result: list[int] = new_stream.to_list()
                assert len(result) <= len(TestStream.TEST_VALUES)
                assert all(value in TestStream.TEST_VALUES for value in result)

        class TestErrorCase:
            def test_none_predicate__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When predicate is None, should raise TypeError exception"""
                # GIVEN test_stream
                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    test_stream.drop_while(None)

    class TestLimit:
        """limit method tests"""

        class TestNominalCase:
            def test_empty_stream__should__return_an_empty_stream(self) -> None:
                """When stream is empty, should return an empty stream"""
                # GIVEN
                stream: Stream[int] = Stream.empty()
                limit: int = 10

                # WHEN
                new_stream: Stream[int] = stream.limit(limit)

                # THEN
                assert new_stream.to_list() == lists.empty_list()

            def test_limit_is_zero__should__return_an_empty_stream(self, test_stream: Stream[int]) -> None:
                """When limit is zero, should return an empty stream"""
                # GIVEN test_stream
                limit: int = 0

                # WHEN
                new_stream: Stream[int] = test_stream.limit(limit)

                # THEN
                assert new_stream.to_list() == lists.empty_list()

            def test_negative_limit__should__return_an_empty_stream(self, test_stream: Stream[int]) -> None:
                """When limit is negative, should return an empty stream"""
                # GIVEN test_stream
                limit: int = -1

                # WHEN
                new_stream: Stream[int] = test_stream.limit(limit)

                # THEN
                assert new_stream.to_list() == lists.empty_list()

            def test_less_values_than_limit__should__return_stream_with_all_elements(
                    self, test_stream: Stream[int]
            ) -> None:
                """When stream contains fewer values than the limit should return a stream with all elements"""
                # GIVEN test_stream
                limit: int = maxsize

                # WHEN
                new_stream: Stream[int] = test_stream.limit(limit)

                # THEN
                result: list[int] = new_stream.to_list()
                assert all(value in result for value in TestStream.TEST_VALUES)

            def test_limit_is_one__should__return_a_truncated_stream_of_one_element(
                    self, test_stream: Stream[int]
            ) -> None:
                """When limit is 1 should return a truncated stream of 1 element"""
                # GIVEN test_stream
                limit: int = 1

                # WHEN
                new_stream: Stream[int] = test_stream.limit(limit)

                # THEN
                result: list[int] = new_stream.to_list()
                assert len(result) == limit
                assert result[0] in TestStream.TEST_VALUES

            def test_limit_exists__should__return_a_truncated_stream(
                    self, test_stream: Stream[int]
            ) -> None:
                """When limit exists should return a truncated stream"""
                # GIVEN test_stream
                limit: int = 20

                # WHEN
                new_stream: Stream[int] = test_stream.limit(limit)

                # THEN
                result: list[int] = new_stream.to_list()
                assert len(result) == limit
                assert all(value in TestStream.TEST_VALUES for value in result)

        class TestErrorCase:
            def test_none_limit__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When limit is None, should raise TypeError exception"""
                # GIVEN test_stream
                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    test_stream.limit(None)

    class TestMax:
        """max method tests"""

        class TestNominalCase:
            def test_empty_stream__should__return_an_empty_joptional(self) -> None:
                """When stream is empty, should return an empty joptional"""
                # GIVEN
                stream: Stream[int] = Stream.empty()

                # WHEN
                result: JOptional[int] = stream.max()

                # THEN
                assert result.is_present() is False

            def test_default_comparator__should__return_a_joptional_with_max_natural_value(
                    self, test_stream: Stream[int]
            ) -> None:
                """When default comparator, should return a joptional describing the natural max value"""
                # GIVEN test_stream
                # WHEN
                result: JOptional[int] = test_stream.max()

                # THEN
                assert result.is_present() is True
                assert result.get() == TestStream.TEST_RANGE_MAX_VALUE

            def test_comparator_exists__should__return_a_joptional_with_max_value_as_described_by_comparator(
                    self, test_stream: Stream[int]
            ) -> None:
                """
                When comparator is provided, should return a joptional describing the max value described by the
                comparator
                """
                # GIVEN test_stream
                reverse_comparator: BiFunction[int, int, int] = lambda i1, i2: 0 if i1 == i2 else 1 if i1 < i2 else -1

                # WHEN
                result: JOptional[int] = test_stream.max(reverse_comparator)

                # THEN
                assert result.is_present() is True
                assert result.get() == TestStream.TEST_RANGE_MIN_VALUE

        class TestErrorCase:
            def test_none_comparator__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When comparator is None, should raise TypeError exception"""
                # GIVEN test_stream
                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    test_stream.max(None)

    class TestMin:
        """min method tests"""

        class TestNominalCase:
            def test_empty_stream__should__return_an_empty_joptional(self) -> None:
                """When stream is empty, should return an empty joptional"""
                # GIVEN
                stream: Stream[int] = Stream.empty()

                # WHEN
                result: JOptional[int] = stream.min()

                # THEN
                assert result.is_present() is False

            def test_default_comparator__should__return_a_joptional_with_min_natural_value(
                    self, test_stream: Stream[int]
            ) -> None:
                """When default comparator, should return a joptional describing the natural min value"""
                # GIVEN test_stream
                # WHEN
                result: JOptional[int] = test_stream.min()

                # THEN
                assert result.is_present() is True
                assert result.get() == TestStream.TEST_RANGE_MIN_VALUE

            def test_comparator_exists__should__return_a_joptional_with_min_value_as_described_by_comparator(
                    self, test_stream: Stream[int]
            ) -> None:
                """
                When comparator is provided, should return a joptional describing the min value described by the
                comparator
                """
                # GIVEN test_stream
                reverse_comparator: BiFunction[int, int, int] = lambda i1, i2: 0 if i1 == i2 else 1 if i1 < i2 else -1

                # WHEN
                result: JOptional[int] = test_stream.min(reverse_comparator)

                # THEN
                assert result.is_present() is True
                assert result.get() == TestStream.TEST_RANGE_MAX_VALUE

        class TestErrorCase:
            def test_none_comparator__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When comparator is None, should raise TypeError exception"""
                # GIVEN test_stream
                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    test_stream.min(None)

    class TestPeek:
        """peek method tests"""

        class TestNominalCase:

            def test_consumer_exists_should__call_consumer_for_each_stream_element(
                    self, test_stream: Stream[int]
            ) -> None:
                with Manager() as manager:
                    # GIVEN test_stream
                    result_list: list[int] = manager.list()
                    consumer: Consumer[int] = result_list.append

                    # WHEN
                    test_stream.peek(consumer)
                    test_stream.to_list()

                    # THEN
                    assert len(result_list) == len(TestStream.TEST_VALUES)
                    assert all(elem in result_list for elem in TestStream.TEST_VALUES)

        class TestErrorCase:
            def test_none_consumer__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When consumer is None, should raise TypeError exception"""
                # GIVEN test_stream
                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    test_stream.peek(None)

    class TestReduce:
        """reduce method tests"""

        @staticmethod
        def _reduce_accumulator(i1: int, i2: int) -> int:
            """test reduce accumulator because lambda cannot be pickled"""
            return i1 + i2

        class TestNominalCase:

            def test_empty_stream__should__return_identity(self) -> None:
                """When stream is empty, should return identity value"""
                # GIVEN
                stream: Stream[int] = Stream.empty()
                identity: int = 0
                accumulator: BiFunction[int, int, int] = TestStream.TestReduce._reduce_accumulator

                # WHEN
                result: int = stream.reduce(identity, accumulator)

                # THEN
                assert result == identity

            def test_stream_with_values__should__return_reduction_result(
                    self, test_stream: Stream[int]
            ) -> None:
                """When stream is empty, should return identity value"""
                # GIVEN
                identity: int = 0
                accumulator: BiFunction[int, int, int] = TestStream.TestReduce._reduce_accumulator

                # WHEN
                result: int = test_stream.reduce(identity, accumulator)

                # THEN
                assert result == sum(TestStream.TEST_VALUES)

        class TestErrorCase:
            def test_none_accumulator__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When consumer is None, should raise TypeError exception"""
                # GIVEN test_stream
                identity: int = 0

                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    test_stream.reduce(identity, None)

    class TestSkip:
        """skip method tests"""

        class TestNominalCase:
            def test_empty_stream__should__return_an_empty_stream(self) -> None:
                """When stream is empty, should return an empty stream"""
                # GIVEN
                stream: Stream[int] = Stream.empty()
                skip: int = 10

                # WHEN
                new_stream: Stream[int] = stream.skip(skip)

                # THEN
                assert new_stream.to_list() == lists.empty_list()

            def test_less_values_than_skip__should__return_an_empty_stream(self, test_stream: Stream[int]) -> None:
                """When stream contains fewer values than skip should return an empty"""
                # GIVEN test_stream
                skip: int = maxsize

                # WHEN
                new_stream: Stream[int] = test_stream.skip(skip)

                # THEN
                assert new_stream.to_list() == lists.empty_list()

            def test_skip_is_zero__should__return_stream_with_all_elements(
                    self, test_stream: Stream[int]
            ) -> None:
                """When skip is zero, should return a stream with all elements"""
                # GIVEN test_stream
                skip: int = 0

                # WHEN
                new_stream: Stream[int] = test_stream.skip(skip)

                # THEN
                result: list[int] = new_stream.to_list()
                assert all(value in result for value in TestStream.TEST_VALUES)

            def test_negative_skip__should__return_an_empty_stream(self, test_stream: Stream[int]) -> None:
                """When skip is negative, should return a stream with all elements"""
                # GIVEN test_stream
                skip: int = -1

                # WHEN
                new_stream: Stream[int] = test_stream.skip(skip)

                # THEN
                result: list[int] = new_stream.to_list()
                assert all(value in result for value in TestStream.TEST_VALUES)

            def test_skip_exists__should__return_a_truncated_stream(
                    self, test_stream: Stream[int]
            ) -> None:
                """When skip exists should return a truncated stream"""
                # GIVEN test_stream
                skip: int = 20

                # WHEN
                new_stream: Stream[int] = test_stream.skip(skip)

                # THEN
                result: list[int] = new_stream.to_list()
                assert len(result) == len(TestStream.TEST_VALUES) - skip
                assert all(value in TestStream.TEST_VALUES for value in result)

        class TestErrorCase:
            def test_none_parameter__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When parameter is None, should raise TypeError exception"""
                # GIVEN test_stream
                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    test_stream.skip(None)

    class TestTakeWhile:
        """take_while method tests"""

        class TestNominalCase:
            def test_empty_stream__should__return_an_empty_stream(self) -> None:
                """When stream is empty, should return an empty stream"""
                # GIVEN
                stream: Stream[int] = Stream.empty()
                min_value: int = 10
                predicate: Predicate[int] = partial(TestStream._predicate_less_than, value=min_value)

                # WHEN
                new_stream: Stream[int] = stream.take_while(predicate)

                # THEN
                assert new_stream.to_list() == lists.empty_list()

            def test_ordered_stream_with_matched_predicate__should__return_stream_with_taken_values(
                    self, test_stream: Stream[int]
            ) -> None:
                """
                When stream is ordered and predicate match at least one value, should return a stream with taken
                values
                """
                # GIVEN
                stream: Stream[int] = test_stream.ordered_execution()
                min_value: int = 10
                predicate: Predicate[int] = partial(TestStream._predicate_less_than, value=min_value)

                # WHEN
                new_stream: Stream[int] = stream.take_while(predicate)

                # THEN
                result: list[int] = new_stream.to_list()
                assert len(result) == min_value
                assert all(value in result for value in range(0, min_value))

            def test_not_matched_predicate__should__return_stream_with_all_values(
                    self, test_stream: Stream[int]
            ) -> None:
                """When no value match, should return an empty stream"""
                # GIVEN
                predicate: Predicate[int] = is_none

                # WHEN
                new_stream: Stream[int] = test_stream.take_while(predicate)

                # THEN
                assert new_stream.to_list() == lists.empty_list()

            def test_unordered_stream_with_valid_predicate__should__return_a_stream_with_some_taken_values(
                    self, test_stream: Stream[int]
            ) -> None:
                """
                When stream is unordered and predicate match at least one value, should return a stream with some
                taken values (the stream may be empty)
                """
                # GIVEN
                stream: Stream[int] = test_stream.unordered()
                min_value: int = 50
                predicate: Predicate[int] = partial(TestStream._predicate_less_than, value=min_value)

                # WHEN
                new_stream: Stream[int] = stream.take_while(predicate)

                # THEN
                result: list[int] = new_stream.to_list()
                assert len(result) <= min_value
                assert all(value in TestStream.TEST_VALUES for value in result)

        class TestErrorCase:
            def test_none_predicate__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When predicate is None, should raise TypeError exception"""
                # GIVEN test_stream
                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    test_stream.take_while(None)

    class TestSorted:
        """sorted method tests"""

        class TestNominalCase:
            def test_empty_stream__should__return_an_empty_ordered_stream(self) -> None:
                """When stream is empty, should return an empty stream with ordered configuration"""
                # GIVEN
                stream: Stream[int] = Stream.empty()

                # WHEN
                new_stream: Stream[int] = stream.sorted()

                # THEN
                assert new_stream.to_list() == lists.empty_list()
                assert new_stream._ordered_execution is True

            def test_default_comparator__should__return_an_ordered_natural_order_sorted_stream(
                    self, parallel_config_fixture: ParallelConfiguration
            ) -> None:
                """When uses default comparator, should return an ordered stream with natural ordered values"""
                # GIVEN
                values: list[int] = [2, 5, 1, 3, 7, 120, 7, 10]
                stream: Stream[int] = Stream.of_iterable(values).parallel(parallel_config_fixture)

                # WHEN
                new_stream: Stream[int] = stream.sorted()

                # THEN
                expected_result: list[int] = sorted(values)
                result: list[int] = new_stream.to_list()
                assert new_stream._ordered_execution is True
                assert result == expected_result

            def test_comparator_exists__should__return_an_ordered_stream_based_on_comparator(
                    self, parallel_config_fixture: ParallelConfiguration
            ) -> None:
                """
                When comparator is provided, should return an ordered stream with ordered values based on given
                comparator
                """
                # GIVEN
                values: list[int] = [2, 5, 1, 3, 7, 120, 7, 10]
                stream: Stream[int] = Stream.of_iterable(values).parallel(parallel_config_fixture)
                reverse_comparator: BiFunction[int, int, int] = lambda i1, i2: 0 if i1 == i2 else 1 if i1 < i2 else -1

                # WHEN
                new_stream: Stream[int] = stream.sorted(reverse_comparator)

                # THEN
                expected_result: list[int] = sorted(values, reverse=True)
                result: list[int] = new_stream.to_list()
                assert new_stream._ordered_execution is True
                assert result == expected_result

        class TestErrorCase:
            def test_none_comparator__should__raise_type_error(self, test_stream: Stream[int]) -> None:
                """When comparator is None, should raise TypeError exception"""
                # GIVEN test_stream
                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    test_stream.sorted(None)

    class TestToGenerator:
        """to_generator method tests"""

        class TestNominalCase:

            def test_empty_stream__should__generate_nothing(self) -> None:
                """When stream is empty, should generate nothing"""
                # GIVEN
                stream: Stream[int] = Stream.empty()

                # WHEN
                generator: Iterator[int] = stream.to_generator()

                # THEN
                with raises(StopIteration):
                    generator.__next__()

            def test_ordered_stream__should__generate_mapped_values_in_source_order(
                    self, test_stream: Stream[int]
            ) -> None:
                """When stream is ordered, should return mapped values in source order"""
                # GIVEN test_stream
                stream: Stream[str] = test_stream.ordered_execution().map(str)

                # WHEN
                generator: Iterator[str] = stream.to_generator()

                # THEN
                for value in TestStream.TEST_STR_VALUES:
                    assert generator.__next__() == value
                with raises(StopIteration):
                    generator.__next__()

            def test_unordered_stream__should__generate_mapped_values_in_random_order(
                    self, test_stream: Stream[int]
            ) -> None:
                # GIVEN test_stream
                stream: Stream[str] = test_stream.unordered().map(str)

                # WHEN
                generator: Iterator[str] = stream.to_generator()

                # THEN
                counter: int = 0
                for value in generator:
                    counter += 1
                    assert value in TestStream.TEST_STR_VALUES
                assert counter == len(TestStream.TEST_STR_VALUES)



# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import islice
from typing import Iterator, Iterable

from dev4py.utils.objects import require_non_none
from dev4py.utils.types import V


def get_chunks(values: Iterable[V], chunksize: int) -> Iterator[list[V]]:
    """
    Returns a generator of chunk where values will be chopped into chunks of size chunksize

    Args:
        values: The values to chunk
        chunksize: The chunk size

    Returns:
        Iterator[list[V]]: The chunk iterator

    Raises:
        TypeError: if values or chunksize is None
        ValueError: if chunksize is negative
    """
    require_non_none(chunksize)
    it = iter(require_non_none(values))
    while True:
        chunk = list(islice(it, chunksize))
        if not chunk:
            return
        yield chunk



"""
The `dicts` module provides a set of utility functions to simplify dict operations
"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import Optional, Any

from dev4py.utils.JOptional import JOptional
from dev4py.utils.objects import non_none, require_non_none, to_none
from dev4py.utils.types import K, V, Supplier


def is_dict(value: Any) -> bool:
    """
    If the given value is a dict, returns true, otherwise false

    Returns:
        bool: true if the given value is a dict, otherwise false
    """
    return isinstance(value, dict)


def _get_value(dictionary: dict[K, V], key: K) -> Optional[V]:
    """
    private function to replace get_joptional_value lambda
    Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
    """
    # lambda d: d.get(key)
    return dictionary.get(key)


def get_joptional_value(dictionary: Optional[dict[K, V]], key: K) -> JOptional[V]:
    """
    Tries to get a value from a dict with the given key and returns a JOptional describing the result

    Args:
        dictionary: The dict
        key: The key

    Returns:
        JOptional[V]: An empty JOptional if dictionary is None or the searched key result is None, otherwise a JOptional
        with a present value
    """
    if non_none(dictionary) and not is_dict(dictionary):
        raise TypeError("Optional[dict[K, V]] parameter is required")

    return JOptional \
        .of_noneable(dictionary) \
        .map(partial(_get_value, key=key))


def get_value(
        dictionary: Optional[dict[K, V]], key: K, default_supplier: Supplier[Optional[V]] = to_none
) -> Optional[V]:
    """
    Returns the value from a dict with the given key if presents, otherwise returns the result produced by the supplying
    function

    Args:
        dictionary: The dict
        key: The searched key
        default_supplier: The supplying function that produces a value to be returned

    Returns:
        Optional[V]: The value, if present, otherwise the result produced by the supplying function (even if dictionary
        is None)

    """
    return get_joptional_value(dictionary, key).or_else_get(default_supplier)


def get_value_from_path(
        dictionary: Optional[dict[K, Any]], path: list[Any], default_supplier: Supplier[Optional[V]] = to_none
) -> Optional[V]:
    """
    Returns the value from a deep dict (dict of dicts) with the given key path if present, otherwise returns the result
    produced by the supplying function

    Args:
        dictionary: The dict
        path: The searched key path
        default_supplier: The supplying function that produces a value to be returned

    Returns:
        Optional[V]: The value, if present, otherwise the result produced by the supplying function (even if dictionary
        is None)

    Raises:
        TypeError: if dictionary is not None and not a dict
    """
    if non_none(dictionary) and not is_dict(dictionary):
        raise TypeError("Optional[dict[K, V]] dictionary parameter must be a dict or None value")

    if not path:
        return default_supplier()

    if len(path) == 1:
        return get_value(dictionary, path[0], default_supplier)

    current_path_value: Any = get_value(dictionary, path[0])
    return get_value_from_path(
        current_path_value,
        path[1:],
        default_supplier
    ) if is_dict(current_path_value) else default_supplier()


def put_value(dictionary: dict[K, V], key: K, value: V) -> Optional[V]:
    """
    Associates the specified value with the specified key in the given map. If the map previously contained a mapping
    for the key, the old value is returned and replaced by the specified value

    Args:
        dictionary: The dict
        key: The key with which the specified value is to be associated
        value: The value to be associated with the specified key

    Returns:
        Optional[V]: The previous value associated with key, or None if there was no mapping for key.

    Raises:
        TypeError: if dictionary is not a dict
    """
    if not is_dict(dictionary):
        raise TypeError("dictionary must be a dict value")

    result: Optional[V] = dictionary.get(key)
    dictionary[key] = value
    return result


def empty_dict() -> dict[Any, Any]:
    """
    Returns an empty dict
    Returns:
        dict[Any, Any]: an empty dict
    """
    return {}


def update(dict_1: dict[K, V], dict_2: dict[K, V]) -> dict[K, V]:
    """
    Adds all elements of the second dict to the first one and returns it

    Args:
        dict_1: The dict where add elements
        dict_2: The dict with elements to add

    Returns:
        dict_1 (dict[K, V]): The first dict with added elements from dict_2

    Raises:
        TypeError: if dict_1 or dict_2 is None
    """
    require_non_none(dict_1).update(require_non_none(dict_2))
    return dict_1



# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

from dev4py.utils.objects import require_non_none
from dev4py.utils.types import T


def empty_list() -> list[Any]:
    """
    Returns an empty list

    Returns:
        list[Any]: An empty list
    """
    return []


def append(lst: list[T], value: T) -> list[T]:
    """
    Adds the given value to the given list and returns the list

    Args:
        lst: the list
        value: the value to add

    Returns:
        list[T]: The list with the added value

    Raises:
        TypeError: if the list is None
    """
    require_non_none(lst).append(value)
    return lst


def extend(lst_1: list[T], lst_2: list[T]) -> list[T]:
    """
    Adds all values from the second list to the first one and returns it

    Args:
        lst_1: The list where to add elements
        lst_2: The list with the elements to add

    Returns:
        lst_1 (list[T]): The first list with added elements

    Raises:
        TypeError: if lst_1 or lst_2 is None
    """
    require_non_none(lst_1).extend(require_non_none(lst_2))
    return lst_1



"""
The `types` module provides a set commonly used types useful for static type checking
"""
# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TypeVar, TypeAlias, Callable, Union, Awaitable, ParamSpec

IN = TypeVar('IN')  # pragma: no mutate
K = TypeVar('K')  # pragma: no mutate
N = TypeVar('N')  # pragma: no mutate
OUT = TypeVar('OUT')  # pragma: no mutate
R = TypeVar('R')  # pragma: no mutate
T = TypeVar('T')  # pragma: no mutate
U = TypeVar('U')  # pragma: no mutate
V = TypeVar('V')  # pragma: no mutate

P = ParamSpec('P')  # pragma: no mutate
# Fix Pycharm false positive warning
P: ParamSpec = P  # pragma: no mutate

# See: https://peps.python.org/pep-0484/#type-aliases
#   Examples:
#     -> Function[int, bool] is equivalent to Callable[[int], bool]
#     -> Predicate[str] is equivalent to Callable[[str], bool]
#     -> Supplier[str] is equivalent to Callable[[], str]
# Note (on march 2022): Use mypy for typing validation (By default pycharm is not compliant)
Function: TypeAlias = Callable[[T], R]
"""Function[T, R]: A function that accepts one T type argument and produces a R type result"""
Function.__doc__ = \
    "Function[T, R]: A function that accepts one T type argument and produces a R type result"  # pragma: no mutate

Predicate: TypeAlias = Callable[[T], bool]
"""Predicate[T]: A predicate (boolean-valued function) of one T type argument"""
Predicate.__doc__ = "Predicate[T]: A predicate (boolean-valued function) of one T type argument"  # pragma: no mutate

Consumer: TypeAlias = Callable[[T], None]
"""Consumer[T]: An operation that accepts a single T type argument and returns no result"""
Consumer.__doc__ = \
    "Consumer[T]: An operation that accepts a single T type argument and returns no result"  # pragma: no mutate

Supplier: TypeAlias = Callable[[], R]
"""Supplier[R]: A supplier of results of R type"""
Supplier.__doc__ = "Supplier[R]: A supplier of results of R type"  # pragma: no mutate

Runnable: TypeAlias = Callable[[], None]
"""A function that accepts no T type argument and returns no result"""
Runnable.__doc__ = "A function that accepts no T type argument and returns no result"  # pragma: no mutate

BiFunction: TypeAlias = Callable[[T, U], R]
"""Function[T, U, R]: A function that accepts two arguments and produces a R type result"""
BiFunction.__doc__ = \
    "Function[T, U, R]: A function that accepts two arguments and produces a R type result"  # pragma: no mutate

SyncOrAsync: TypeAlias = Union[Awaitable[T], T]
"""SyncOrAsync[T]: is used to specify that a value can be sync or async"""
SyncOrAsync.__doc__ = \
    "SyncOrAsync[T]: is used to specify that a value can be sync or async"  # pragma: no mutate

BiConsumer: TypeAlias = Callable[[T, U], None]
"""BiConsumer[T, U]: An operation that accepts two arguments and produces no result"""  # pragma: no mutate
Consumer.__doc__ = \
    "BiConsumer[T, U]: An operation that accepts two arguments and produces no result"  # pragma: no mutate
